from policyengine_us.model_api import *


class or_cdcc_relevant_expenses(Variable):
    value_type = float
    entity = TaxUnit
    label = "Oregon working family household and dependent care expenses"
    unit = USD
    definition_period = YEAR
    reference = "https://oregon.public.law/statutes/ors_315.264"
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        p_cdcc = parameters(period).gov.irs.credits.cdcc
        p_or = parameters(period).gov.states["or"].tax.income.credits.wfhdc

        # First, cap based on the number of eligible care receivers
        expenses = tax_unit("tax_unit_childcare_expenses", period)
        capped_eligible_people = tax_unit("capped_count_cdcc_eligible", period)

        total_max_amount = p_or.cap * capped_eligible_people
        capped_eligible_expenses = min_(expenses, total_max_amount)

        return min_(
            capped_eligible_expenses,
            tax_unit("min_head_spouse_earned", period),
        )



from policyengine_us.model_api import *


class ri_retirement_income_subtraction_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Rhode Island retirement income subtraction"
    definition_period = YEAR
    reference = "http://webserver.rilin.state.ri.us/Statutes/title44/44-30/44-30-12.HTM"
    # Eligibility is the same as Social Security's eligibility
    # https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/Social%20Security%20Worksheet_w.pdf
    # STEP 1: Eligibility
    defined_for = StateCode.RI

    def formula(tax_unit, period, parameters):
        income = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        older_spouse_birth_year = tax_unit("older_spouse_birth_year", period)

        p_ss = parameters(
            period
        ).gov.states.ri.tax.income.agi.subtractions.social_security.limit
        age_eligible = older_spouse_birth_year <= p_ss.birth_year

        # Income eligibility.
        p_tri = parameters(
            period
        ).gov.states.ri.tax.income.agi.subtractions.taxable_retirement_income
        income_eligible = income < p_tri.income_limit[filing_status]

        return age_eligible & income_eligible



from policyengine_us.variables.household.demographic.geographic.state_code import (
    StateCode,
)
from functools import reduce
from policyengine_core.model_api import *
from policyengine_us.entities import *
from policyengine_us.tools.general import *
from pathlib import Path
from policyengine_us.typing import *
import warnings
from policyengine_us.tools.cloning import get_stored_variables

warnings.filterwarnings("ignore")

REPO = Path(__file__).parent


def all_of_variables(variables: List[str]) -> Formula:
    def formula(entity, period, parameters):
        value = True
        for variable in variables:
            value = value & (add(entity, period, [variable]) > 0)
        return value

    return formula


STATES = [
    "AL",
    "AK",
    "AZ",
    "AR",
    "CA",
    "CO",
    "CT",
    "DC",
    "DE",
    "FL",
    "GA",
    "HI",
    "ID",
    "IL",
    "IN",
    "IA",
    "KS",
    "KY",
    "LA",
    "ME",
    "MD",
    "MA",
    "MI",
    "MN",
    "MS",
    "MO",
    "MT",
    "NE",
    "NV",
    "NH",
    "NJ",
    "NM",
    "NY",
    "NC",
    "ND",
    "OH",
    "OK",
    "OR",
    "PA",
    "RI",
    "SC",
    "SD",
    "TN",
    "TX",
    "UT",
    "VT",
    "VA",
    "WA",
    "WV",
    "WI",
    "WY",
    "PR",
]



from policyengine_us.model_api import *


class nj_taking_property_tax_deduction(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Household taking New Jersey property tax deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/new-jersey/2022/title-54a/section-54a-3a-17/"
    defined_for = "nj_property_tax_deduction_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.nj.tax.income.main

        # follows NJ-1040 form Worksheet H
        deduction = tax_unit("nj_potential_property_tax_deduction", period)

        # calculate taxes without deduction
        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        taxable_income_before_deduction = tax_unit(
            "nj_taxable_income_before_property_tax_deduction", period
        )
        taxes_without_deduction = select(
            [
                filing_status == status.SINGLE,
                filing_status == status.JOINT,
                filing_status == status.HEAD_OF_HOUSEHOLD,
                filing_status == status.SURVIVING_SPOUSE,
                filing_status == status.SEPARATE,
            ],
            [
                p.single.calc(taxable_income_before_deduction),
                p.joint.calc(taxable_income_before_deduction),
                p.head_of_household.calc(taxable_income_before_deduction),
                p.surviving_spouse.calc(taxable_income_before_deduction),
                p.separate.calc(taxable_income_before_deduction),
            ],
        )

        # calculate taxes with deduction
        taxable_income_after_deduction = max_(
            0, taxable_income_before_deduction - deduction
        )
        taxes_with_deduction = select(
            [
                filing_status == status.SINGLE,
                filing_status == status.JOINT,
                filing_status == status.HEAD_OF_HOUSEHOLD,
                filing_status == status.SURVIVING_SPOUSE,
                filing_status == status.SEPARATE,
            ],
            [
                p.single.calc(taxable_income_after_deduction),
                p.joint.calc(taxable_income_after_deduction),
                p.head_of_household.calc(taxable_income_after_deduction),
                p.surviving_spouse.calc(taxable_income_after_deduction),
                p.separate.calc(taxable_income_after_deduction),
            ],
        )

        # calculate credit amount
        credit_amount = parameters(
            period
        ).gov.states.nj.tax.income.credits.property_tax.amount
        separate = filing_status == status.SEPARATE
        cohabitating = tax_unit("cohabitating_spouses", period)
        credit_amount = credit_amount / (1 + separate * cohabitating)

        # choose between taxable income deduction and refundable credit
        return (taxes_without_deduction - taxes_with_deduction) > credit_amount



from policyengine_us.model_api import *


class tx_ottanf_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Texas One-Time TANF (OTTANF)"
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-2421-eligibility-criteria",
        "https://www.law.cornell.edu/regulations/texas/1-Tex-Admin-Code-SS-372-801",
    )
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        # OTTANF eligibility per A-2421:
        # 1. Must meet all regular TANF eligibility requirements
        #    (child, income, resources, immigration)
        # 2. Must be eligible for TANF grant of $10 or more
        # 3. Must meet one of four crisis criteria

        # Regular TANF eligibility (child, income, resources, immigration)
        regular_tanf_eligible = spm_unit("tx_tanf_eligible", period)

        # Must be eligible for at least $10 grant
        payment_standard = spm_unit("tx_tanf_payment_standard", period)
        countable_income = spm_unit("tx_tanf_countable_income", period)
        potential_grant = payment_standard - countable_income

        p = parameters(period).gov.states.tx.tanf
        eligible_for_min_grant = potential_grant >= p.minimum_grant

        # Crisis criteria (user input)
        crisis = spm_unit("tx_ottanf_crisis_criteria", period)

        return regular_tanf_eligible & eligible_for_min_grant & crisis



from policyengine_us.model_api import *


def create_ctc_older_child_supplement() -> Reform:
    class ctc_child_individual_maximum(Variable):
        value_type = float
        entity = Person
        label = "CTC maximum amount (child)"
        unit = USD
        documentation = (
            "The CTC entitlement in respect of this person as a child."
        )
        definition_period = YEAR
        reference = (
            "https://www.law.cornell.edu/uscode/text/26/24#a",
            "https://www.law.cornell.edu/uscode/text/26/24#h",
            "https://www.law.cornell.edu/uscode/text/26/24#i",
        )
        defined_for = "is_tax_unit_dependent"

        def formula(person, period, parameters):
            age = person("age", period)
            base_amount = parameters(period).gov.irs.credits.ctc.amount.base
            base_amount = base_amount.calc(age)
            oldest_child = person("child_index", period) == 1
            p_ref = parameters(period).gov.contrib.ctc.oldest_child_supplement
            oldest_child_supplement = oldest_child * p_ref.amount
            return base_amount + oldest_child_supplement

    class reform(Reform):
        def apply(self):
            self.update_variable(ctc_child_individual_maximum)

    return reform


def create_ctc_older_child_supplement_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_ctc_older_child_supplement()

    p = parameters(period).gov.contrib.ctc.oldest_child_supplement

    if p.in_effect:
        return create_ctc_older_child_supplement()
    else:
        return None


ctc_older_child_supplement = create_ctc_older_child_supplement_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class nm_cdcc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = (
        "Eligible household for the New Mexico dependent child day care credit"
    )
    definition_period = YEAR
    reference = "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503752/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgDsfAEwBKADTJspQhACKiQrgCe0AOSapEQmFwJlqjdt37DIAMp5SAIQ0AlAKIAZZwDUAggDkAws5SpGAARtCk7BISQA"
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        # Filer can not be a dependent on another tax return
        dependent_elsewhere = tax_unit("head_is_dependent_elsewhere", period)
        p = parameters(period).gov.states.nm.tax.income.credits.cdcc
        # Filer has to be be gainfully employed to receive credit
        has_earnings = person("earned_income", period) > 0
        # If a joint return is filed, both spouses must be gainfully employed
        # unless one person is disabled
        filing_status = tax_unit("filing_status", period)
        joint = filing_status == filing_status.possible_values.JOINT
        head = person("is_tax_unit_head", period)
        spouse = person("is_tax_unit_spouse", period)
        head_has_earnings = tax_unit.any(head & has_earnings)
        spouse_has_earnings = tax_unit.any(spouse & has_earnings)
        both_have_earnings = head_has_earnings & spouse_has_earnings
        disabled = person("is_disabled", period)
        disabled_eligible = tax_unit.any((head | spouse) & disabled)
        joint_eligible = both_have_earnings | disabled_eligible
        employment_eligible = where(joint, joint_eligible, head_has_earnings)
        # Filer can not receive tanf to be eligible
        receives_tanf = tax_unit.spm_unit("tanf", period) > 0
        # Filers have to have modified gross income at or below a limit
        nm_modified_gross_income = tax_unit("nm_modified_gross_income", period)
        income_limit = (
            parameters(period).gov.dol.minimum_wage
            * p.income_limit_as_fraction_of_minimum_wage
            * WEEKS_IN_YEAR
            * p.full_time_hours
        )
        income_eligible = nm_modified_gross_income <= income_limit
        return (
            ~dependent_elsewhere
            & employment_eligible
            & ~receives_tanf
            & income_eligible
        )



from policyengine_us.model_api import *


class ca_capi_countable_vehicle_value(Variable):
    value_type = float
    unit = USD
    entity = SPMUnit
    label = "California CAPI countable vehicle value"
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "https://www.cdss.ca.gov/Portals/9/CAPI/CAPI_Regulations-Accessible.pdf"

    def formula(spm_unit, period, parameters):
        vehicle_value = spm_unit.household("household_vehicles_value", period)
        p = parameters(period).gov.states.ca.cdss.capi.resources.vehicle
        return max_(vehicle_value - p.disregard, 0)



from policyengine_us.model_api import *


class oh_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio Earned Income Credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://codes.ohio.gov/ohio-revised-code/section-5747.71",
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf#page=21",
    )
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        match = parameters(period).gov.states.oh.tax.income.credits.eitc.rate
        return federal_eitc * match



from policyengine_us.model_api import *


class meets_ccdf_activity_test(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = YEAR
    documentation = "Indicates whether parent or parents meet activity test (working/in job training/in educational program)"
    label = "Activity test for CCDF"



from policyengine_us.model_api import *


class il_tanf_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = (
        "Eligible for Illinois Temporary Assistance for Needy Families (TANF)"
    )
    definition_period = MONTH
    reference = "https://www.dhs.state.il.us/page.aspx?item=30358"
    defined_for = StateCode.IL

    def formula(spm_unit, period, parameters):
        non_financial_eligible = spm_unit(
            "il_tanf_non_financial_eligible", period
        )
        income_eligible = spm_unit("il_tanf_income_eligible", period)

        return non_financial_eligible & income_eligible



from policyengine_us.model_api import *


class PellGrantFormula(Enum):
    A = "A"
    B = "B"
    C = "C"


class pell_grant_formula(Variable):
    value_type = Enum
    possible_values = PellGrantFormula
    default_value = PellGrantFormula.A
    entity = Person
    label = "Pell Grant formula"
    definition_period = YEAR

    def formula(person, period, parameters):
        has_dependents = person.tax_unit("tax_unit_dependents", period) > 0
        is_head = person("is_tax_unit_head", period)
        is_spouse = person("is_tax_unit_spouse", period)
        head_or_spouse = is_head | is_spouse
        return select(
            [
                has_dependents & ~head_or_spouse,
                ~has_dependents & head_or_spouse,
            ],
            [PellGrantFormula.A, PellGrantFormula.B],
            # C formula applies for `has_dependents & head_or_spouse`.
            # Technically this also catches `~has_dependents & ~head_or_spouse`
            # but that's not a valid combination.
            default=PellGrantFormula.C,
        )



from policyengine_us.model_api import *


class wv_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia taxable income"
    unit = USD
    definition_period = YEAR
    reference = "https://code.wvlegislature.gov/11-21-4E/"
    defined_for = StateCode.WV

    def formula(tax_unit, period, parameters):
        agi = tax_unit("wv_agi", period)
        low_income_exclusion = tax_unit(
            "wv_low_income_earned_income_exclusion", period
        )
        exemptions = tax_unit("wv_personal_exemption", period)
        return max_(agi - low_income_exclusion - exemptions, 0)



from policyengine_us.model_api import *


class il_aabd_earned_income_after_exemption_person(Variable):
    value_type = float
    entity = Person
    definition_period = MONTH
    label = "Illinois Aid to the Aged, Blind or Disabled (AABD) earned income after exemption per person"
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-113.120",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        p = parameters(period).gov.states.il.dhs.aabd.income.exemption
        gross_earned_income = person("il_aabd_gross_earned_income", period)
        expense_exemption = person("il_aabd_expense_exemption_person", period)
        adjusted_income = max_(gross_earned_income - expense_exemption, 0)
        flat_exemption_excess = person(
            "il_aabd_flat_exemption_excess_over_unearned_income", period
        )
        income_after_flat_exemption = max_(
            adjusted_income - flat_exemption_excess, 0
        )

        blind = person("is_blind", period)
        exemption = where(
            blind,
            p.blind.calc(income_after_flat_exemption),
            p.elderly_or_disabled.calc(income_after_flat_exemption),
        )
        return income_after_flat_exemption - exemption



from policyengine_us.model_api import *


class snap_earned_income_deduction(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    documentation = (
        "Earned income deduction for calculating SNAP benefit amount"
    )
    label = "SNAP earned income deduction"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e_2"

    def formula(spm_unit, period, parameters):
        deduction_rate = parameters(
            period
        ).gov.usda.snap.income.deductions.earned_income
        return spm_unit("snap_earned_income", period) * deduction_rate



from policyengine_us.model_api import *


class ca_state_supplement_aged_blind_disabled_amount(Variable):
    value_type = float
    entity = SPMUnit
    label = "California SSI state supplement aged disabled and blind amount"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.CA
    reference = "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=WIC&sectionNum=12200"

    def formula(spm_unit, period, parameters):
        blind_amount = spm_unit("ca_state_supplement_blind_amount", period)
        aged_disabled_amount = spm_unit(
            "ca_state_supplement_aged_disabled_amount", period
        )
        return max_(blind_amount, aged_disabled_amount)



from policyengine_us.model_api import *


class state_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "State standard deduction"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        # States that adopt the federal standard deduction
        # Based on comments in state_standard_deductions.yaml
        FEDERAL_STANDARD_DEDUCTION_STATES = [
            "CT",  # Connecticut
            "ID",  # Idaho
            "ME",  # Maine
            "MO",  # Missouri
            "ND",  # North Dakota
            "NM",  # New Mexico
            "SC",  # South Carolina
            "UT",  # Utah
            "CO",  # Colorado
        ]

        # Get the sum of state-specific standard deductions
        state_specific = add(
            tax_unit,
            period,
            parameters(period).gov.states.household.state_standard_deductions,
        )
        state_code = tax_unit.household("state_code_str", period)

        # Special handling for states that need individual vs joint standard deduction maximum
        STATES_WITH_INDIVIDUAL_JOINT_MAXIMUM = {
            "MT": {
                "indiv": "mt_standard_deduction_indiv",
                "joint": "mt_standard_deduction_joint",
            },
            "IA": {
                "indiv": "ia_standard_deduction_indiv",
                "joint": "ia_standard_deduction_joint",
            },
            "DE": {
                "indiv": "de_standard_deduction_indv",
                "joint": "de_standard_deduction_joint",
            },
            "AR": {
                "indiv": "ar_standard_deduction_indiv",
                "joint": "ar_standard_deduction_joint",
            },
            "MS": {
                "indiv": "ms_standard_deduction_indiv",
                "joint": "ms_standard_deduction_joint",
            },
            "KY": {
                "indiv": "ky_standard_deduction_indiv",
                "joint": "ky_standard_deduction_joint",
            },
        }

        # Calculate maximum standard deductions for applicable states
        for state, variables in STATES_WITH_INDIVIDUAL_JOINT_MAXIMUM.items():
            is_state = state_code == state
            indiv_deductions = add(tax_unit, period, [variables["indiv"]])
            joint_deductions = add(tax_unit, period, [variables["joint"]])
            max_deductions = max_(indiv_deductions, joint_deductions)
            state_specific = where(is_state, max_deductions, state_specific)

        # Check if the state adopts federal standard deduction
        uses_federal = np.isin(state_code, FEDERAL_STANDARD_DEDUCTION_STATES)

        # Get federal standard deduction
        federal_deduction = tax_unit("standard_deduction", period)

        # Return federal deduction for states that adopt it, otherwise state-specific
        return where(uses_federal, federal_deduction, state_specific)



from policyengine_us.model_api import *


class or_wfhdc_household_size_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Household size eligible for the Oregon working family household and dependent care credit"
    documentation = "Oregon Working Family Household and Dependent Care Credit household eligibility"
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/schedule-or-wfhdc-inst_101-195-1_2022.pdf#pahe=1",
        "https://law.justia.com/codes/oregon/2021/volume-08/chapter-315/section-315-264/",
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states["or"].tax.income.credits.wfhdc
        # Over two individuals have to be present in the tax unit.
        return tax_unit("tax_unit_size", period) >= p.min_tax_unit_size



from policyengine_us.model_api import *


class ma_tafdc_work_related_expense_deduction(Variable):
    value_type = float
    unit = USD
    entity = Person
    label = "Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) work-related expense deduction"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-270"
    )
    defined_for = StateCode.MA

    adds = ["gov.states.ma.dta.tcap.deductions.work_related_expenses.amount"]



from policyengine_us.model_api import *


class vt_percentage_capital_gains_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont percentage capital gains exclusion"
    unit = USD
    documentation = "This can be selected to be subtracted from federal adjusted gross income in Vermont as percentage capital gains exclusion."
    definition_period = YEAR
    defined_for = StateCode.VT
    reference = (
        "https://tax.vermont.gov/sites/tax/files/documents/IN-153-2022.pdf#page=1"  # 2022 Schedule IN-153 Vermont Capital Gains Exclusion Calculation
        "https://legislature.vermont.gov/statutes/section/32/151/05811"  # Titl. 32 V.S.A. § 5811(21)(B)(ii)
        "https://tax.vermont.gov/sites/tax/files/documents/IN-153%20Instr-2022.pdf"
    )

    def formula(tax_unit, period, parameters):
        # Get adjusted net capital gain, which is capped at 0
        adjusted_net_capital_gain = tax_unit(
            "adjusted_net_capital_gain", period
        )
        qualified_dividend_income = add(
            tax_unit, period, ["qualified_dividend_income"]
        )
        reduced_adjusted_net_capital_gain = max_(
            adjusted_net_capital_gain - qualified_dividend_income, 0
        )
        p = parameters(
            period
        ).gov.states.vt.tax.income.agi.exclusions.capital_gain
        # The percentage exclusion equals to a percentage of
        # the adjusted net capital gain and has a maximum value
        percentage_exclusion = (
            reduced_adjusted_net_capital_gain * p.percentage.rate
        )
        return min_(percentage_exclusion, p.percentage.cap)



from policyengine_us.model_api import *


class homeowners_association_fees(Variable):
    value_type = float
    entity = SPMUnit
    label = "Homeowners association fees"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class is_chip_eligible_pregnant(Variable):
    value_type = bool
    entity = Person
    label = "Pregnant person eligible for CHIP"
    documentation = "Determines if a pregnant person is eligible for the Children's Health Insurance Program through either the standard pregnant pathway or FCEP"
    definition_period = YEAR
    reference = (
        "https://www.ssa.gov/OP_Home/ssact/title21/2110.htm",
        "https://www.medicaid.gov/medicaid/national-medicaid-chip-program-information/medicaid-childrens-health-insurance-program-basic-health-program-eligibility-levels",
        "https://www.kff.org/affordable-care-act/state-indicator/medicaid-and-chip-income-eligibility-limits-for-pregnant-women-as-a-percent-of-the-federal-poverty-level",
    )

    def formula(person, period, parameters):
        # A person is eligible for CHIP as pregnant if they qualify through either
        # the standard pregnant pathway or the FCEP pathway

        # Check standard pregnant pathway eligibility
        standard_eligible = person(
            "is_chip_eligible_standard_pregnant_person", period
        )

        # Check FCEP pathway eligibility
        fcep_eligible = person("is_chip_fcep_eligible_person", period)

        # Eligible if either pathway qualifies
        return standard_eligible | fcep_eligible



from policyengine_us.model_api import *


class ky_pension_income_exclusion_exemption_eligible(Variable):
    value_type = bool
    entity = Person
    label = "KY Pension Income Exclusion Exemption Eligible"
    definition_period = YEAR
    defined_for = StateCode.KY
    reference = "https://revenue.ky.gov/Forms/Schedule%20P%202022.pdf"

    # Determine if person is eligible for exemption.
    # True if they retired from federal or KY government or receive railroad benefits.
    adds = [
        "retired_from_federal_government",
        "retired_from_ky_government",
        "railroad_benefits",
    ]



from policyengine_us.model_api import *


class mt_taxable_income_joint(Variable):
    value_type = float
    entity = Person
    label = "Montana taxable income when married couples file jointly"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT
    reference = (
        "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/05/Montana-Idividiual-Income-Tax-Return-Form-2-2022v6.2.pdf#page=1",
        "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2022/12/Form-2-2022-Instructions.pdf#page=16",
    )

    def formula(person, period, parameters):
        is_head = person("is_tax_unit_head", period)

        # For joint filers, use mt_agi_joint which pools income and subtractions
        # at tax unit level before applying them. This ensures subtractions from
        # one spouse can offset income from the other spouse.
        total_agi = person.tax_unit("mt_agi_joint", period)

        standard_deduction = add(
            person.tax_unit, period, ["mt_standard_deduction_joint"]
        )
        itemized_deductions = add(
            person.tax_unit, period, ["mt_itemized_deductions_joint"]
        )
        # Tax units can claim the larger of the itemized or standard deductions
        deductions = max_(itemized_deductions, standard_deduction)
        exemptions = add(
            person.tax_unit,
            period,
            ["mt_personal_exemptions_joint", "mt_dependent_exemptions_person"],
        )

        return is_head * max_(0, total_agi - deductions - exemptions)



from policyengine_us.model_api import *


class is_wic_eligible(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    documentation = "Is eligible for the Special Supplemental Nutrition Program for Women, Infants and Children (WIC)"
    label = "Is eligible for WIC"
    reference = "https://www.law.cornell.edu/cfr/text/7/246.7#c_1"

    def formula(person, period, parameters):
        meets_income_test = person.spm_unit("meets_wic_income_test", period)
        meets_categorical_test = person(
            "meets_wic_categorical_eligibility", period
        )
        nutritional_risk = person("is_wic_at_nutritional_risk", period)
        return (meets_income_test | meets_categorical_test) & nutritional_risk



from policyengine_us.model_api import *


class ky_additions(Variable):
    value_type = float
    entity = Person
    label = "Kentucky additions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://apps.legislature.ky.gov/law/statutes/statute.aspx?id=53498"
        "https://revenue.ky.gov/Forms/Schedule%20M%202022.pdf#page=1"
        "https://revenue.ky.gov/Forms/740%20Packet%20Instructions%205-9-23.pdf#page=23"
        "https://revenue.ky.gov/Forms/Schedule%20M-2021.pdf#page=1"
        "https://taxsim.nber.org/historical_state_tax_forms/KY/2021/Form%20740%20Packet%20Instructions-2021.pdf#page=27"
    )
    defined_for = StateCode.KY



from policyengine_us.model_api import *


class nm_eitc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for New Mexico EITC"
    definition_period = YEAR
    reference = "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/856ebf4b-3814-49dd-8631-ebe579d6a42b/Personal%20Income%20Tax.pdf"  # 7-2-18.15
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        # Use federal EITC eligibility, but replace demographic eligibility
        # with NM version.
        eitc = parameters.gov.irs.credits.eitc(period)
        investment_income_eligible = tax_unit(
            "eitc_investment_income_eligible", period
        )
        demographic_eligible = tax_unit("nm_eitc_demographic_eligible", period)
        # Define eligibility before considering separate filer limitation.
        eligible = demographic_eligible & investment_income_eligible
        # This parameter is true if separate filers are eligible.
        if eitc.eligibility.separate_filer:
            return eligible
        # If separate filers are not eligible, check if the filer is separate.
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        return eligible & ~separate



from policyengine_us.model_api import *


class va_additions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia additions to federal adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/",
        "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf#page=24",
    )
    defined_for = StateCode.VA



from policyengine_us.model_api import *
from policyengine_us.variables.gov.irs.credits.cdcc.count_cdcc_eligible import (
    count_cdcc_eligible,
)


class ma_dependent_care_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA dependent care credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.mass.gov/info-details/mass-general-laws-c62-ss-6"  # (y)
    )
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ma.tax.income.credits.dependent_care
        # Expenses capped by number of qualifying individuals.
        expenses = tax_unit("tax_unit_childcare_expenses", period)
        count_cdcc_eligible = min_(
            tax_unit("count_cdcc_eligible", period), p.dependent_cap
        )
        # Line 1.
        capped_expenses = min_(expenses, p.amount * count_cdcc_eligible)
        # Skip lines 2 and 3, which are done in intermediate steps.
        # Line 4: Smallest of lines 1, 2, 3.
        amount_if_eligible = min_(
            capped_expenses, tax_unit("min_head_spouse_earned", period)
        )
        # Skip line 5 for prior-year expenses.
        # Married filing separate are ineligible.
        filing_status = tax_unit("filing_status", period)
        eligible = filing_status != filing_status.possible_values.SEPARATE
        return eligible * amount_if_eligible



from policyengine_us.model_api import *


class pr_is_tax_unit_dependent(Variable):
    value_type = bool
    entity = Person
    label = "Puerto Rico tax unit dependent"
    definition_period = YEAR
    defined_for = StateCode.PR
    reference = "https://hacienda.pr.gov/sites/default/files/inst_individuals_2023.pdf#page=28"

    def formula(person, period, parameters):
        # dependent definition in Puerto Rico:
        p = parameters(
            period
        ).gov.territories.pr.tax.income.taxable_income.exemptions.dependent.age_threshold
        age = person("age", period)

        # taxpayer is <21 years old OR >= 65 years old
        age_eligible = (age < p.younger) | (age >= p.older)

        # OR the taxpayer's mother or father
        parent_eligible = person("is_parent_of_filer_or_spouse", period)

        # OR a person over 21 that is incapable of self-support because of a physical or mental condition
        incapable_of_self_support_eligible = person(
            "is_incapable_of_self_care", period
        )

        is_blind = person("is_blind", period)

        # OR a university student <26 years old that attended at least 1 years of an accredited uni
        student_eligible = (person("is_full_time_college_student", period)) & (
            age < p.student
        )
        return (
            student_eligible
            | parent_eligible
            | incapable_of_self_support_eligible
            | is_blind
            | age_eligible
        )



from policyengine_us.model_api import *


class oh_unreimbursed_medical_care_expense_deduction_person(Variable):
    value_type = float
    entity = Person
    label = "Ohio unreimbursed medical and health care expense deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf#page=18",  # Line 36
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf#page=27",
        "https://codes.ohio.gov/ohio-revised-code/section-5747.01",  # R.C. 5747.01(10)
    )
    defined_for = StateCode.OH

    adds = [
        "oh_insured_unreimbursed_medical_care_expenses",
        "oh_uninsured_unreimbursed_medical_care_expenses",
        "long_term_health_insurance_premiums",
    ]



from policyengine_us.model_api import *


class ky_tax_unit_itemizes(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Whether the tax unit in Kentucky itemizes the deductions when married filing separately"
    definition_period = YEAR
    defined_for = StateCode.KY

    def formula(tax_unit, period, parameters):
        itemized_ded = add(tax_unit, period, ["ky_itemized_deductions_indiv"])
        standard_ded = add(tax_unit, period, ["ky_standard_deduction_indiv"])
        return itemized_ded > standard_ded



from policyengine_us.model_api import *


class spm_unit_federal_tax(Variable):
    value_type = float
    entity = SPMUnit
    label = "Federal income tax"
    definition_period = YEAR
    unit = USD

    def formula(spm_unit, period, parameters):
        return sum_contained_tax_units("income_tax", spm_unit, period)



from policyengine_us.model_api import *


class nm_medical_expense_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico unreimbursed medical expense care exemption"
    definition_period = YEAR
    reference = "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503680/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgBsADh4BKADTJspQhACKiQrgCe0AOSapEQmFwJlqjdt37DIAMp5SAIQ0AlAKIAZZwDUAggDkAws5SpGAARtCk7BISQA"
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        p = parameters(
            period
        ).gov.states.nm.tax.income.exemptions.unreimbursed_medical_care_expense
        age = person("age", period)
        medical_expense = add(
            tax_unit, period, ["medical_out_of_pocket_expenses"]
        )
        age_eligible = tax_unit.any(age >= p.age_eligibility)
        expense_eligible = medical_expense >= p.min_expenses
        eligible = age_eligible & expense_eligible
        # Exemption is halved for married filing separately
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        denominator = where(separate, 2, 1)
        numerator = eligible * p.amount
        return numerator / denominator



from policyengine_us.model_api import *


class ca_sf_wftc(Variable):
    value_type = float
    entity = TaxUnit
    label = "San Francisco Working Families Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.sfhsa.org/sites/default/files/media/document/2024-01/form_wfc_english_1.26.24.pdf#page=4"
    defined_for = StateCode.CA

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.local.ca.sf.wftc
        county = tax_unit.household("county_str", period)
        in_sf = county == "SAN_FRANCISCO_COUNTY_CA"
        eitc_eligible = tax_unit("eitc_eligible", period)
        ca_eitc_eligible = tax_unit("ca_eitc_eligible", period)
        eligibility = in_sf & (eitc_eligible | ca_eitc_eligible)
        return p.amount * eligibility



from policyengine_us.model_api import *


class fuel_cell_property_capacity(Variable):
    value_type = float
    entity = TaxUnit
    label = "Capacity of purchased fuel cells"
    documentation = (
        "Kilowatts of capacity of qualified fuel cell properties purchased."
    )
    unit = "kW"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25D#b_1"



from policyengine_us.model_api import *


class flat_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Flat tax"
    unit = USD
    documentation = "Flat income tax on federal AGI or gross income."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.contrib.ubi_center.flat_tax
        # Gross income flat tax.
        gross_income = add(tax_unit, period, ["positive_gross_income"])
        gross_income_flat_tax = gross_income * p.rate.gross_income
        # AGI flat tax.
        filing_status = tax_unit("filing_status", period)
        exemption = p.exemption.agi[filing_status]
        agi = tax_unit("positive_agi", period)
        excess_agi = max_(agi - exemption, 0)
        agi_flat_tax = excess_agi * p.rate.agi
        return gross_income_flat_tax + agi_flat_tax



from policyengine_us.model_api import *


class co_ccap_base_parent_fee(Variable):
    value_type = float
    entity = SPMUnit
    label = "Colorado Child Care Assistance Program base parent fee"
    reference = (
        "https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11042&fileName=8%20CCR%201403-1#page=41",
        "https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11042&fileName=8%20CCR%201403-1#page=62",
    )
    unit = USD
    definition_period = MONTH

    def formula(spm_unit, period, parameters):
        year = period.start.year
        if period.start.month >= 10:
            instant_str = f"{year}-10-01"
        else:
            instant_str = f"{year - 1}-10-01"
        p = parameters(instant_str).gov.states.co.ccap
        # Calculate base parent fee and add on parent fee.
        gross_income = spm_unit("co_ccap_countable_income", period)
        # snap_fpg is monthly.
        snap_fpg = spm_unit("snap_fpg", period)
        # Calculate base parent fee scaled (note income is monthly):
        # When income_scaled <= 1: income_scaled * 0.01
        # When income_scaled > 1: [1 * 0.01 + (income_scaled - 1) * 0.14]
        # Multiply by fpg afterward to scale back up
        gross_income_fpg_ratio = gross_income / snap_fpg
        base_parent_fee_scaled = p.parent_fee.base.calc(gross_income_fpg_ratio)
        return base_parent_fee_scaled * snap_fpg



from policyengine_us.model_api import *


class taxsim_sprofinc(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        "SSTB income for the spouse of the taxpayer (TAXSIM). Assumed zero."
    )
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class tanf_person(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Per-capita TANF"
    documentation = (
        "Per-capita value of Temporary Assistance for Needy Families benefit."
    )
    unit = USD

    def formula(person, period, parameters):
        return person.spm_unit("tanf", period) / person.spm_unit.nb_persons()



from policyengine_us.model_api import *


class hi_alternative_tax_on_capital_gains(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii alternative tax on capital gains"
    unit = USD
    definition_period = YEAR
    defined_for = "hi_alternative_tax_on_capital_gains_eligible"

    def formula(tax_unit, period, parameters):
        # The alternative tax on capital is computed as the sum of the eligible taxable income
        # multiplied by the normal tax rate or by a flat tax rate.
        eligible_taxable_income = tax_unit(
            "hi_taxable_income_for_alternative_tax", period
        )
        p = parameters(period).gov.states.hi.tax.income
        filing_status = tax_unit("filing_status", period)
        statuses = filing_status.possible_values
        normal_tax_rate = select(
            [
                filing_status == statuses.SINGLE,
                filing_status == statuses.SEPARATE,
                filing_status == statuses.JOINT,
                filing_status == statuses.SURVIVING_SPOUSE,
                filing_status == statuses.HEAD_OF_HOUSEHOLD,
            ],
            [
                p.rates.single.calc(eligible_taxable_income),
                p.rates.separate.calc(eligible_taxable_income),
                p.rates.joint.calc(eligible_taxable_income),
                p.rates.surviving_spouse.calc(eligible_taxable_income),
                p.rates.head_of_household.calc(eligible_taxable_income),
            ],
        )
        taxable_income = tax_unit("hi_taxable_income", period)
        excess_taxable_income = max_(
            0, taxable_income - eligible_taxable_income
        )
        flat_tax_rate = p.alternative_tax.rate * excess_taxable_income
        return flat_tax_rate + normal_tax_rate



from policyengine_us.model_api import *


class or_wfhdc_employment_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Employment eligible for the Oregon working family household and dependent care credit"
    documentation = "Oregon Working Family Household and Dependent Care Credit household eligibility"
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/schedule-or-wfhdc-inst_101-195-1_2022.pdf#pahe=1",
        "https://law.justia.com/codes/oregon/2021/volume-08/chapter-315/section-315-264/",
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        # 1) you are working
        earned_income = person("earned_income", period)
        head = person("is_tax_unit_head", period)
        earned_income_eligible = tax_unit.any(head * earned_income) > 0
        # 2)  you are single and you attended school (full-time or part-time)
        filing_status = tax_unit("filing_status", period)
        attend_school = tax_unit.any(head & person("is_in_k12_school", period))
        single_student = (
            filing_status == filing_status.possible_values.SINGLE
        ) & attend_school
        # 3) you are married filing jointly and one spouse attended school (full-time) or was disabled
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        is_full_time_student = person("is_full_time_student", period)
        married_eligible = tax_unit.any(
            is_full_time_student & is_head_or_spouse
        )
        joint_head_or_spouse_student = (
            filing_status == filing_status.possible_values.JOINT
        ) & married_eligible

        return (
            earned_income_eligible
            | single_student
            | joint_head_or_spouse_student
        )



from policyengine_us.model_api import *


class ca_riv_general_relief_net_earned_income(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "Riverside County GA net earned income after deductions"
    definition_period = MONTH

    def formula(person, period, parameters):
        earned = person("ca_riv_general_relief_earned_income", period)
        deductions = person(
            "ca_riv_general_relief_earned_income_deductions", period
        )

        return max_(earned - deductions, 0)



from policyengine_us.model_api import *


class vt_normal_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont normal income tax before non-refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VT
    reference = (
        "https://casetext.com/statute/vermont-statutes/title-32-taxation-and-finance/chapter-151-income-taxes/subchapter-002-taxation-of-individuals-trusts-and-estates/section-5822-tax-on-income-of-individuals-estates-and-trusts",  # Vermont §5822. Tax on income of individuals, estates, and trusts (a)
        "https://tax.vermont.gov/sites/tax/files/documents/RateSched-2021.pdf#page=1",  # Vermont 2021 Income Tax Rate Schedules
        "https://tax.vermont.gov/sites/tax/files/documents/RateSched-2022.pdf#page=1",  # Vermont 2022 Income Tax Return Booklet Forms and Instructions
    )

    def formula(tax_unit, period, parameters):
        income = tax_unit("vt_taxable_income", period)
        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        p = parameters(period).gov.states.vt.tax.income.rates
        return select(
            [
                filing_status == status.SINGLE,
                filing_status == status.JOINT,
                filing_status == status.SEPARATE,
                filing_status == status.SURVIVING_SPOUSE,
                filing_status == status.HEAD_OF_HOUSEHOLD,
            ],
            [
                p.single.calc(income),
                p.joint.calc(income),
                p.separate.calc(income),
                p.surviving_spouse.calc(income),
                p.head_of_household.calc(income),
            ],
        )



from policyengine_us.model_api import *


class ky_personal_tax_credits_indiv(Variable):
    value_type = float
    entity = Person
    label = (
        "Kentucky personal tax credits when married couples file separately"
    )
    unit = USD
    reference = "https://apps.legislature.ky.gov/law/statutes/statute.aspx?id=53500#page=3"  # (3) (a)
    definition_period = YEAR
    defined_for = StateCode.KY

    adds = [
        "ky_blind_personal_tax_credits",
        "ky_aged_personal_tax_credits",
        "ky_military_personal_tax_credits",
    ]



from policyengine_us.model_api import *


class nm_aged_blind_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico aged and blind exemption"
    unit = USD
    definition_period = YEAR
    reference = (
        # 7-2-5.2. EXEMPTION--INCOME OF PERSONS SIXTY-FIVE AND OLDER OR BLIND Page 22
        "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/856ebf4b-3814-49dd-8631-ebe579d6a42b/Personal%20Income%20Tax.pdf#page=34",
        # Tax Form Instructions Page ADJ-5 TABLE 1. Exemptions for Persons 65 or Older or Blind
        "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf#page=50",
        # 7-2-5.2. Exemption; income of persons sixty-five and older or blind
        "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503666/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgBsogJQAaZNlKEIARUSFcAT2gBydRIiEwuBIuVrN23fpABlPKQBCagEoBRADKOAagEEAcgGFHE0jAAI2hSdjExIA",
    )
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.nm.tax.income.exemptions.blind_and_aged
        filing_status = tax_unit("filing_status", period)
        statuses = filing_status.possible_values
        agi = tax_unit("adjusted_gross_income", period)
        age_threshold = p.age_threshold

        # Check if taxpayer is blind or aged.
        blind_head = tax_unit("blind_head", period)
        aged_head = tax_unit("age_head", period) >= age_threshold

        # Check if taxpayer is eligible.
        # New Mexico does not double exemptions if the same individual is both aged and blind.
        head_eligible = blind_head | aged_head

        # Check if spouse is blind or aged.
        blind_spouse = tax_unit("blind_spouse", period)
        aged_spouse = tax_unit("age_spouse", period) >= age_threshold

        # Check if spouse is eligible.
        spouse_eligible = blind_spouse | aged_spouse

        eligible_count = head_eligible.astype(int) + spouse_eligible.astype(
            int
        )

        # Use `right=True` to reflect "over ... but not over ...".
        amount = select(
            [
                filing_status == statuses.SINGLE,
                filing_status == statuses.JOINT,
                filing_status == statuses.HEAD_OF_HOUSEHOLD,
                filing_status == statuses.SEPARATE,
                filing_status == statuses.SURVIVING_SPOUSE,
            ],
            [
                p.single.calc(agi, right=True),
                p.joint.calc(agi, right=True),
                p.head_of_household.calc(agi, right=True),
                p.separate.calc(agi, right=True),
                p.surviving_spouse.calc(agi, right=True),
            ],
        )
        return eligible_count * amount



from policyengine_us.model_api import *


class adjusted_gross_income_person(Variable):
    value_type = float
    entity = Person
    label = "Federal adjusted gross income for each person"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/62"

    def formula(person, period, parameters):
        gross_income = person("irs_gross_income", period)
        # calculate ald sums by person
        PERSON_ALDS = [
            "self_employment_tax_ald",
            "self_employed_health_insurance_ald",
            "self_employed_pension_contribution_ald",
        ]
        person_ald_vars = [f"{ald}_person" for ald in PERSON_ALDS]
        ald_sum_person = add(person, period, person_ald_vars)
        # split other alds evenly between head and spouse
        all_alds = parameters(period).gov.irs.ald.deductions
        other_alds = list(set(all_alds) - set(PERSON_ALDS))
        ald_sum_taxunit = add(person.tax_unit, period, other_alds)
        is_head = person("is_tax_unit_head", period)
        is_spouse = person("is_tax_unit_spouse", period)
        fstatus = person.tax_unit("filing_status", period)
        frac = where(fstatus == fstatus.possible_values.JOINT, 0.5, 1.0)
        ald_sum_taxunit_shared = (is_head | is_spouse) * ald_sum_taxunit * frac
        # calculate AGI by person
        agi = gross_income - ald_sum_person - ald_sum_taxunit_shared
        if parameters(period).gov.contrib.ubi_center.basic_income.taxable:
            basic_income = person.tax_unit("basic_income", period)
            # split basic income evenly between head and spouse
            basic_income_shared = (is_head | is_spouse) * basic_income * frac
            agi += basic_income_shared
        return agi



from policyengine_us.model_api import *


class ma_liheap_standard_payment(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts LIHEAP standard payment"
    definition_period = YEAR
    defined_for = "ma_liheap_eligible"
    reference = "https://www.mass.gov/doc/fy-2025-heap-income-eligibility-benefit-chart-may-8-2025/download"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.ma.doer.liheap.standard.amount
        utility_category = spm_unit("ma_liheap_utility_category", period)
        benefit_level = spm_unit("ma_liheap_benefit_level", period)
        is_subsidized = spm_unit("receives_housing_assistance", period)

        return where(
            is_subsidized,
            p.subsidized[benefit_level][utility_category],
            p.non_subsidized[benefit_level][utility_category],
        )



from policyengine_us.model_api import *


class ny_itemizes(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Itemizes New York deductions"
    documentation = "Tax units who itemize their federal deductions can opt to itemize their New York deductions. However, if a standard deduction causes a lower tax liability, they must choose that."
    definition_period = YEAR
    defined_for = StateCode.NY
    reference = "https://www.nysenate.gov/legislation/laws/TAX/613"

    def formula(tax_unit, period, parameters):
        federal_itemizes = tax_unit("tax_unit_itemizes", period)
        standard_deduction = tax_unit("ny_standard_deduction", period)
        itemized_deductions = tax_unit("ny_itemized_deductions", period)
        return federal_itemizes & (itemized_deductions >= standard_deduction)



from policyengine_us.model_api import *


class taxpayer_has_itin(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Tax unit head or spouse has ITIN"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        has_itin = person("has_itin", period)
        return tax_unit.any(head_or_spouse & has_itin)



from policyengine_us.model_api import *


class wi_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin income tax before credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1f.pdf"
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1-Inst.pdf"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1f.pdf"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1-Inst.pdf"
        "https://docs.legis.wisconsin.gov/misc/lfb/informational_papers/january_2023/0002_individual_income_tax_informational_paper_2.pdf"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        taxinc = tax_unit("wi_taxable_income", period)
        fstatus = tax_unit("filing_status", period)
        p = parameters(period).gov.states.wi.tax.income
        return select(
            [
                fstatus == fstatus.possible_values.SINGLE,
                fstatus == fstatus.possible_values.JOINT,
                fstatus == fstatus.possible_values.SURVIVING_SPOUSE,
                fstatus == fstatus.possible_values.SEPARATE,
                fstatus == fstatus.possible_values.HEAD_OF_HOUSEHOLD,
            ],
            [
                p.rates.single.calc(taxinc),
                p.rates.joint.calc(taxinc),
                p.rates.joint.calc(taxinc),
                p.rates.separate.calc(taxinc),
                p.rates.head_of_household.calc(taxinc),
            ],
        )



from policyengine_us.model_api import *


class home_mortgage_interest(Variable):
    value_type = float
    entity = Person
    label = "Interest paid on a home mortgage"
    unit = USD
    definition_period = YEAR
    documentation = "Home mortgage interest, including both reported and not reported on federal Form 1098."



from policyengine_us.model_api import *


class ut_federal_deductions_for_taxpayer_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah federal deductions considered for taxpayer credit"
    unit = USD
    documentation = "These federal deductions are added to the Utah personal exemption to determine the Utah taxpayer credit."
    definition_period = YEAR
    defined_for = StateCode.UT
    reference = "https://le.utah.gov/xcode/Title59/Chapter10/59-10-S114.html?v=C59-10-S114_2022032320220323"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions
        filing_status = tax_unit("filing_status", period)

        us_itemizing = tax_unit("tax_unit_itemizes", period)
        std_ded = tax_unit("standard_deduction", period)

        # Subtract SALT from Itemized Deductions
        deductions = [
            deduction
            for deduction in p.itemized_deductions
            if deduction not in ["salt_deduction"]
        ]
        item_ded = add(tax_unit, period, deductions)

        # Include Real Estate Taxes in Itemized Deductions
        real_estate_tax = add(tax_unit, period, ["real_estate_taxes"])

        capped_real_estate_tax = min_(
            real_estate_tax, p.itemized.salt_and_real_estate.cap[filing_status]
        )

        # Line 12. Federal Standard or Itemized Deductions
        total_item_ded = item_ded + capped_real_estate_tax
        return where(us_itemizing, total_item_ded, std_ded)



from policyengine_us.model_api import *


class ca_agi_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "CA AGI subtractions from federal AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ftb.ca.gov/forms/2021/2021-540-ca-instructions.html"
        "https://www.ftb.ca.gov/forms/2022/2022-540-ca-instructions.html"
    )
    defined_for = StateCode.CA

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ca.tax.income.agi
        total_subtractions = add(tax_unit, period, p.subtractions)
        # Prevent negative subtractions from acting as additions
        return max_(0, total_subtractions)



from policyengine_us.model_api import *


class ri_property_tax_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Rhode Island property tax credit eligibility status"
    definition_period = YEAR
    reference = (
        "http://webserver.rilin.state.ri.us/Statutes/TITLE44/44-33/44-33-3.htm",  # (1) & (2) determines age and disability eligibility
        "https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-01/2021-ri-1040h_w.pdf#page=1",
    )
    defined_for = StateCode.RI

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ri.tax.income.credits.property_tax
        # minimum age eligibility
        greater_head_or_spouse_age = tax_unit(
            "greater_age_head_spouse", period
        )
        age_eligible = greater_head_or_spouse_age >= p.age_threshold
        # disability eligibility
        head_is_disabled = tax_unit("head_is_disabled", period)
        spouse_is_disabled = tax_unit("spouse_is_disabled", period)
        head_or_spouse_disabled = head_is_disabled | spouse_is_disabled
        household_income = tax_unit("ri_property_tax_household_income", period)
        income_threshold = p.rate.one_person.thresholds[-1]
        # The tax form RI-1040H specifies the income of a household as a eligibility requirement
        household_income_eligible = household_income <= income_threshold
        return (
            age_eligible | head_or_spouse_disabled
        ) & household_income_eligible



from policyengine_us.model_api import *


class la_military_pay_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana military pay exclusion"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.LA
    reference = "http://www.legis.la.gov/legis/Law.aspx?d=101760"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.la.tax.income.exempt_income.military_pay_exclusion
        military_pay = add(tax_unit, period, ["military_service_income"])
        return min_(p.max_amount, military_pay)



from policyengine_us.model_api import *


class ca_tanf_resources_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = (
        "Eligible for the California CalWORKs based on the available resources"
    )
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FCalWORKs%2FCalWORKs%2F42-200_Property%2F42-200_Property.htm%23Policybc-2&rhtocid=_3_1_2_0_1"

    def formula(spm_unit, period, parameters):
        resources = spm_unit("ca_tanf_resources", period)
        limit = spm_unit("ca_tanf_resources_limit", period)
        return resources <= limit



from policyengine_us.model_api import *


class md_montgomery_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Montgomery County, Maryland EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://www3.montgomerycountymd.gov/311/Solutions.aspx?SolutionId=1-4DAM0I"
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.local.md.montgomery.tax.income.credits.eitc.refundable

        # Check whether unit is in montgomery county
        county = tax_unit.household("county_str", period)
        in_montgomery = county == "MONTGOMERY_COUNTY_MD"

        # state EITC
        state_refundable_eitc = tax_unit("md_refundable_eitc", period)

        # apply county match
        return state_refundable_eitc * p.match * in_montgomery



from policyengine_us.model_api import *


class il_aabd_person(Variable):
    value_type = float
    entity = Person
    label = "Illinois Aid to the Aged, Blind or Disabled (AABD) cash benefit per person"
    unit = USD
    definition_period = MONTH
    defined_for = "il_aabd_eligible_person"

    def formula(person, period, parameters):
        total_needs = person("il_aabd_need_standard_person", period)
        countable_income = person("il_aabd_countable_income", period)
        return max_(total_needs - countable_income, 0)



from policyengine_us.model_api import *


class va_agi_less_exemptions_person(Variable):
    value_type = float
    entity = Person
    label = "Difference between individual VAGI and personal exemption amounts"
    defined_for = StateCode.VA
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf#page=19"

    def formula(person, period, parameters):
        # Get the individual's VAGI.
        va_agi = person("va_agi_person", period)

        # Get the individual's personal exemption amount.
        personal_exemption = person("va_personal_exemption_person", period)

        # Get the individual's aged/blind exemption amount.
        aged_blind_exemption = person("va_aged_blind_exemption_person", period)

        # Subtract the exemptions from the VAGI.
        return max_(va_agi - personal_exemption - aged_blind_exemption, 0)



from policyengine_us.model_api import *


class TxTanfCaretakerType(Enum):
    NON_CARETAKER = "Non-caretaker"
    CARETAKER_WITHOUT_SECOND_PARENT = "Caretaker without second parent"
    CARETAKER_WITH_SECOND_PARENT = "Caretaker with second parent"


class tx_tanf_caretaker_type(Variable):
    value_type = Enum
    possible_values = TxTanfCaretakerType
    default_value = TxTanfCaretakerType.NON_CARETAKER
    entity = SPMUnit
    label = "Texas TANF caretaker type"
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/c-110-tanf"
    )
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        # Caretaker type determines payment standard amount
        # Based on number of parents included in certified group

        person = spm_unit.members
        eligible_parent = person("tx_tanf_eligible_parent", period)
        eligible_child = person("tx_tanf_eligible_child", period)

        # Count payment-eligible parents and children
        parent_count = spm_unit.sum(eligible_parent)
        has_eligible_child = spm_unit.any(eligible_child)

        # Determine caretaker type
        # Non-caretaker: No parents in certified group (child-only cases)
        # Caretaker without second: One parent in certified group
        # Caretaker with second: Two or more parents in certified group
        non_caretaker = has_eligible_child & (parent_count == 0)
        caretaker_without_second = has_eligible_child & (parent_count == 1)
        caretaker_with_second = has_eligible_child & (parent_count >= 2)

        return select(
            [non_caretaker, caretaker_without_second, caretaker_with_second],
            [
                TxTanfCaretakerType.NON_CARETAKER,
                TxTanfCaretakerType.CARETAKER_WITHOUT_SECOND_PARENT,
                TxTanfCaretakerType.CARETAKER_WITH_SECOND_PARENT,
            ],
            default=TxTanfCaretakerType.NON_CARETAKER,
        )



from policyengine_us.model_api import *


class hud_ttp_adjusted_income_share(Variable):
    value_type = float
    entity = SPMUnit
    label = "HUD adjusted income share for total tenant payment"
    unit = USD
    documentation = "HUD adjusted income share for Total Tenant Payment"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/24/5.628#a_2"

    def formula(spm_unit, period, parameters):
        share = parameters(
            period
        ).gov.hud.total_tenant_payment.adjusted_income_share
        income = spm_unit("hud_adjusted_income", period)
        return share * income



from policyengine_us.model_api import *


class snap_standard_utility_allowance(Variable):
    value_type = float
    entity = SPMUnit
    label = "SNAP Standard Utility Allowance"
    unit = USD
    documentation = "The standard utility allowance deduction for SNAP"
    definition_period = MONTH

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.usda.snap.income.deductions.utility.standard
        allowance_type = spm_unit("snap_utility_allowance_type", period)
        allowance_types = allowance_type.possible_values
        region = spm_unit.household("snap_utility_region_str", period)
        spm_size = spm_unit("spm_unit_size", period)
        MAX_SPM_SIZE = 10
        capped_size = min_(MAX_SPM_SIZE, spm_size)
        sua_household_size_dependent = spm_unit(
            "snap_standard_utility_allowance_by_household_size", period
        )

        sua = where(sua_household_size_dependent, 0, p.main[region])

        # change the state code to NC for the states that do not depend on household size to prevent key error
        region = where(sua_household_size_dependent, region, "NC")

        sua = where(
            sua_household_size_dependent,
            p.by_household_size.amount[region][capped_size],
            sua,
        )

        return where(allowance_type == allowance_types.SUA, sua, 0)



from policyengine_us.model_api import *


class la_general_relief_disability_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = MONTH
    label = "Eligible for the Los Angeles County General Relief based on the disability requirements"
    # Person has to be disabled AND unable to work
    # which is curently not implemented
    defined_for = "is_disabled"
    reference = "https://drive.google.com/file/d/1Oc7UuRFxJj-eDwTeox92PtmRVGnG9RjW/view?usp=sharing"



from policyengine_us.model_api import *


class new_clean_vehicle_battery_components_made_in_north_america(Variable):
    value_type = float
    entity = TaxUnit
    label = "Percent of new clean vehicle's battery components made in North America"
    documentation = "Percent of newly purchased new clean vehicle's battery components (by value) manufactured or assembled in North America"
    unit = "/1"
    definition_period = YEAR



from policyengine_us.model_api import *


class deductible_interest_expense(Variable):
    value_type = float
    entity = Person
    label = "Interest paid on all loans"
    unit = USD
    definition_period = YEAR

    adds = ["deductible_mortgage_interest", "non_mortgage_interest"]



from policyengine_us.model_api import *


class co_ccap_entry_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the entry of Colorado Child Care Assistance Program through income"
    reference = "https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11042&fileName=8%20CCR%201403-1#page=19"
    definition_period = MONTH
    # defined_for = StateCode.CO

    def formula(spm_unit, period, parameters):
        fpg_eligible = spm_unit("co_ccap_fpg_eligible", period)
        smi_eligible = spm_unit("co_ccap_smi_eligible", period)
        return fpg_eligible & smi_eligible



from policyengine_us.model_api import *


class ca_amti(Variable):
    value_type = float
    entity = TaxUnit
    label = "California alternative minimum taxable income"
    defined_for = StateCode.CA
    unit = USD
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/forms/2022/2022-540-p.pdf"

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.ca.tax.income.amt.exemption
        p_us = parameters(period).gov.irs.income.amt.capital_gains

        amti_before_ded = tax_unit("ca_pre_exemption_amti", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        # Calculation from Schedule 540 P Line 21 Separate calculation
        # line 1 - total amti
        # Line 2
        maximum_exemption = p.amti.threshold.upper[filing_status]
        # Line 3
        reduced_amti = max_(amti_before_ded - maximum_exemption, 0)
        # Line 4
        reduced_amti_rate = reduced_amti * p_us.capital_gain_excess_tax_rate
        # Line 5
        separate_amti_calc = min_(reduced_amti_rate, p.amount[filing_status])

        # line 21
        separate_amti_eligible = separate & (reduced_amti > 0)
        separate_amti = separate_amti_eligible * separate_amti_calc
        return amti_before_ded + separate_amti



from policyengine_us.model_api import *


class mo_taxable_income(Variable):
    value_type = float
    entity = Person
    label = "Missouri AGI minus deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-A_2021.pdf"
        "https://dor.mo.gov/forms/MO-1040%20Instructions_2021.pdf#page=8"
        "https://dor.mo.gov/forms/MO-1040%20Instructions_2022.pdf#page=8"
        "https://www.revisor.mo.gov/main/OneSection.aspx?section=143.111&bid=7201&hl="
    )
    defined_for = StateCode.MO

    def formula(person, period, parameters):
        # calculate tax unit MO AGI
        tax_unit = person.tax_unit
        mo_agi = person("mo_adjusted_gross_income", period)
        unit_mo_agi = tax_unit.sum(mo_agi)

        # calculate sum of all tax unit MO deductions
        #   2021 and 2022 Form MO-1040 instructions on page 8 say:
        #     If you claimed the standard deduction on your federal return,
        #     enter the standard deduction amount for your filing status.
        #     The amounts are listed on Form MO-1040, Line 14.
        #     If you itemized on your federal return, you may want to itemize
        #     on your Missouri return or take the standard deduction, whichever
        #     results in a higher deduction.
        federal_itemizer = tax_unit("tax_unit_itemizes", period)
        itmded = tax_unit("mo_itemized_deductions", period)
        stdded = tax_unit("standard_deduction", period)  # same as federal
        mo_deduction = where(federal_itemizer, max_(itmded, stdded), stdded)
        unit_mo_deductions = (
            mo_deduction
            + tax_unit("mo_federal_income_tax_deduction", period)
            + tax_unit("mo_pension_and_ss_or_ssd_deduction", period)
        )
        # Note: There would also be a personal and/or dependent exemptions
        # as part of this formula, but they are legally based on eligibility
        # for the federal versions of those exemptions, both of which are
        # suspended through 2025 federally.

        # calculate taxable income for tax unit
        unit_taxinc = max_(0, unit_mo_agi - unit_mo_deductions)

        # allocate tax unit taxable income by each individual's share of
        # unit AGI (use a mask rather than where to avoid a divide-by-zero
        # warning with default share value being zero)
        ind_agi_share = np.zeros_like(unit_mo_agi)
        mask = unit_mo_agi > 0
        ind_agi_share[mask] = mo_agi[mask] / unit_mo_agi[mask]

        return ind_agi_share * unit_taxinc



from policyengine_us.model_api import *


class meets_tanf_non_cash_net_income_test(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets net income test for TANF non-cash benefit"
    documentation = "Income eligibility (net income as a percent of the poverty line) for TANF non-cash benefit for SNAP BBCE"
    definition_period = MONTH

    def formula(spm_unit, period, parameters):
        # Determine if the net income limit applies to the household.
        applies = parameters(
            period
        ).gov.hhs.tanf.non_cash.income_limit.net_applies
        state = spm_unit.household("state_code_str", period)
        # Varies depending on if the household has elderly and disabled people.
        hheod = spm_unit("is_tanf_non_cash_hheod", period)
        net_limit_applies = where(
            hheod, applies.hheod[state], applies.non_hheod[state]
        ).astype(bool)
        # All limits and incomes here expressed as % of FPG.
        net_income = spm_unit("snap_net_income_fpg_ratio", period)
        net_limit = parameters(period).gov.usda.snap.income.limit.net
        # Either the net limit doesn't apply or they pass it.
        return ~net_limit_applies | (net_income <= net_limit)



from policyengine_us.model_api import *


class eitc_relevant_investment_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "EITC-relevant investment income"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        no_loss_capital_gains = max_(0, tax_unit("net_capital_gains", period))
        return (
            add(
                tax_unit,
                period,
                ["net_investment_income", "tax_exempt_interest_income"],
            )
            # replace limited-loss capital gains with no-loss capital gains
            - tax_unit("loss_limited_net_capital_gains", period)
            + no_loss_capital_gains
        )



from policyengine_us.model_api import *


class ca_riv_general_relief(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Riverside County General Relief"
    definition_period = MONTH
    defined_for = "ca_riv_general_relief_eligible"

    adds = [
        "ca_riv_general_relief_needs_standards",
        "ca_riv_general_relief_special_needs_amount",
    ]
    subtracts = ["ca_riv_general_relief_countable_income"]



from policyengine_us.model_api import *


class ny_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY taxable income"
    unit = USD
    documentation = "NY AGI less taxable income deductions"
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/611"
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        agi = tax_unit("ny_agi", period)
        deductions = tax_unit("ny_deductions", period)
        exemptions = tax_unit("ny_exemptions", period)
        return max_(0, agi - deductions - exemptions)



from policyengine_us.model_api import *


class WICCategory(Enum):
    PREGNANT = "Pregnant"
    POSTPARTUM = "Postpartum"
    BREASTFEEDING = "Breastfeeding"
    INFANT = "Infant"
    CHILD = "Child"
    NONE = "None"


class wic_category(Variable):
    value_type = Enum
    entity = Person
    definition_period = YEAR
    possible_values = WICCategory
    default_value = WICCategory.NONE
    documentation = "Demographic category for the Special Supplemental Nutrition Program for Women, Infants and Children (WIC)"
    label = "WIC demographic category"
    reference = "https://www.law.cornell.edu/uscode/text/42/1786#b"

    def formula(person, period, parameters):
        pregnant = person("is_pregnant", period)
        mother = person("is_mother", period)
        breastfeeding = person("is_breastfeeding", period)
        age = person("age", period)
        # Categorize mothers based on the minimum age of children in the SPM unit.
        min_age_family = person.family.min(age)
        return select(
            *zip(
                (pregnant, WICCategory.PREGNANT),
                (
                    mother & breastfeeding & (min_age_family < 1),
                    WICCategory.BREASTFEEDING,
                ),
                (mother & (min_age_family < 0.5), WICCategory.POSTPARTUM),
                (age < 1, WICCategory.INFANT),
                (age < 5, WICCategory.CHILD),
                (True, WICCategory.NONE),
            )
        )



from policyengine_us.model_api import *


class ky_files_separately(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Married couple file separately on the Kentucky tax return"
    definition_period = YEAR
    reference = "https://revenue.ky.gov/Forms/740%20Packet%20Instructions%205-9-23.pdf#page=11"
    defined_for = StateCode.KY

    def formula(tax_unit, period, parameters):
        itax_indiv = add(
            tax_unit,
            period,
            ["ky_income_tax_before_non_refundable_credits_indiv"],
        )

        itax_joint = add(
            tax_unit,
            period,
            ["ky_income_tax_before_non_refundable_credits_joint"],
        )
        return itax_indiv < itax_joint



from policyengine_us.model_api import *


class ca_riv_general_relief_ineligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Ineligible person for the Riverside County General Relief"
    definition_period = MONTH
    defined_for = "in_riv"

    def formula(person, period, parameters):
        receives_ssi_or_tanf = (
            add(person.spm_unit, period, ["ssi", "ca_tanf"]) > 0
        )
        immigration_status_eligible = person(
            "ca_riv_general_relief_immigration_status_eligible", period
        )
        meets_work_requirement = person(
            "ca_riv_general_relief_meets_work_requirements", period
        )
        return (
            receives_ssi_or_tanf
            | ~immigration_status_eligible
            | ~meets_work_requirement
        )



from policyengine_us.model_api import *


class disabled_tax_unit_head_or_spouse(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Head or Spouse of tax unit is disabled"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        disabled_person = person("is_disabled", period)
        return tax_unit.any(disabled_person & is_head_or_spouse)



from policyengine_us.model_api import *


class oh_withheld_income_tax(Variable):
    value_type = float
    entity = Person
    label = "Ohio withheld income tax"
    defined_for = StateCode.OH
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        agi = person("adjusted_gross_income_person", period)
        p = parameters(period).gov.states.oh.tax.income
        exempt = p.agi_threshold > agi
        # Since Ohio does not have a standard deduction, we apply the
        # personal exemption amount based on employment income as opposed to AGI
        personal_exemptions = p.exemptions.personal.amount.calc(agi)
        reduced_agi = max_(agi - personal_exemptions, 0)
        return p.rates.calc(reduced_agi) * ~exempt



from policyengine_us.model_api import *


class oh_529_plan_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio deduction for contributions to 529 plans"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=18",
    )
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.oh.tax.income.deductions.plan_529_contributions
        contribution_amount = tax_unit("investment_in_529_plan", period)
        beneficiaries = add(
            tax_unit, period, ["count_529_contribution_beneficiaries"]
        )
        cap = p.cap * beneficiaries
        return min_(contribution_amount, cap)



from policyengine_us.model_api import *


class wv_taxable_property_value(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia taxable property value"
    defined_for = StateCode.WV
    unit = USD
    definition_period = YEAR
    reference = "https://code.wvlegislature.gov/11-21-21/"

    adds = ["assessed_property_value"]
    subtracts = ["wv_homestead_exemption"]



from policyengine_us.model_api import *


class ma_part_a_gross_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Part A gross income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/info-details/mass-general-laws-c62-ss-2"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        dividends = add(tax_unit, period, ["dividend_income"])
        nonnegative_short_term_capital_gains = max_(
            0, add(tax_unit, period, ["short_term_capital_gains"])
        )
        return dividends + nonnegative_short_term_capital_gains



from policyengine_us.model_api import *
from policyengine_core.simulations import Simulation


class meets_ssi_resource_test(Variable):
    value_type = bool
    entity = Person
    label = "Meets SSI resource test"
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        # Assign individuals SSI pass rate probabilistically in microsimulation.
        # Apply policy logic in individual simulation.
        p = parameters(period).gov.ssa.ssi
        if person.simulation.dataset is not None:
            return random(person) < p.eligibility.resources.pass_rate
        joint_claim = person("ssi_claim_is_joint", period)
        personal_resources = person("ssi_countable_resources", period)
        countable_resources = where(
            joint_claim,
            person.marital_unit.sum(personal_resources),
            personal_resources,
        )
        resource_limit = where(
            joint_claim,
            p.eligibility.resources.limit.couple,
            p.eligibility.resources.limit.individual,
        )
        return countable_resources <= resource_limit



from pathlib import Path
from policyengine_core.taxbenefitsystems import TaxBenefitSystem
from policyengine_us.entities import *
from policyengine_us.parameters.gov.irs.uprating import (
    set_irs_uprating_parameter,
)
from policyengine_core.simulations import (
    Simulation as CoreSimulation,
    Microsimulation as CoreMicrosimulation,
    IndividualSim as CoreIndividualSim,
)
from policyengine_us.variables.household.demographic.geographic.state.in_state import (
    create_50_state_variables,
)
from policyengine_us.tools.parameters import backdate_parameters
from policyengine_us.reforms import create_structural_reforms_from_parameters
from policyengine_core.parameters.operations.homogenize_parameters import (
    homogenize_parameter_structures,
)
from policyengine_core.parameters.operations.interpolate_parameters import (
    interpolate_parameters,
)
from policyengine_core.parameters.operations.propagate_parameter_metadata import (
    propagate_parameter_metadata,
)
from policyengine_core.parameters.operations.uprate_parameters import (
    uprate_parameters,
)
from .tools.default_uprating import add_default_uprating

from typing import Annotated


COUNTRY_DIR = Path(__file__).parent

CURRENT_YEAR = 2024
DEFAULT_START_DATE = str(CURRENT_YEAR) + "-01-01"

DEFAULT_DATASET = "hf://policyengine/policyengine-us-data/enhanced_cps_2024.h5"


class CountryTaxBenefitSystem(TaxBenefitSystem):
    """
    The tax-benefit system for the United States.
    This structure is a modification of the -core
    package's base TaxBenefitSystem class.

    Args:
        reform (tuple | None): A tuple of reforms to apply to the system.
        If no reform is applied, the system will be initialized with the
        default tax/benefit parameters.

        start_instant(str: ISO date format YYYY-MM-DD): Optional; The date
        at which the simulation begins; defaults to 2024-01-01; this is a
        temporary patch for structural reforms, and must be set to the start
        date of a structural reform parameter if it begins on a date other
        than the first day of the current year.
    """

    variables_dir = COUNTRY_DIR / "variables"
    auto_carry_over_input_variables = True
    basic_inputs = [
        "state_name",
        "employment_income",
        "age",
    ]
    modelled_policies = COUNTRY_DIR / "modelled_policies.yaml"

    def __init__(
        self,
        reform: tuple | None = None,
        start_instant: Annotated[
            str, "ISO date format YYYY-MM-DD"
        ] = DEFAULT_START_DATE,
    ):
        super().__init__(entities, reform=reform)
        self.load_parameters(COUNTRY_DIR / "parameters")
        self.add_abolition_parameters()
        if reform:
            self.apply_reform_set(reform)
        self.parameters = set_irs_uprating_parameter(self.parameters)
        self.parameters = homogenize_parameter_structures(
            self.parameters, self.variables
        )
        self.parameters = propagate_parameter_metadata(self.parameters)
        self.parameters = interpolate_parameters(self.parameters)
        self.parameters = uprate_parameters(self.parameters)
        self.parameters = propagate_parameter_metadata(self.parameters)
        add_default_uprating(self)

        structural_reform = create_structural_reforms_from_parameters(
            self.parameters, start_instant
        )
        if reform is None:
            reform = ()
        reform = (reform, structural_reform)

        self.parameters = backdate_parameters(
            self.parameters, first_instant="2015-01-01"
        )

        for parameter in self.parameters.get_descendants():
            parameter.modified = False

        if reform is not None:
            self.apply_reform_set(reform)

        self.add_variables(*create_50_state_variables())


system = CountryTaxBenefitSystem()


class Simulation(CoreSimulation):
    """
    A simulation of the tax-benefit system for the United States,
    defined against the base simulation class in the -core package.

    This simulation is commonly used for household-level impacts, as it
    does not include society-wide microdata.

    Args:
        start_instant(str: ISO date format YYYY-MM-DD): Optional; The date
        at which the simulation begins; defaults to 2024-01-01; this is a
        temporary patch for structural reforms, and must be set to the start
        date of a structural reform parameter if it begins on a date other
        than the first day of the current year.
    """

    default_tax_benefit_system = CountryTaxBenefitSystem
    default_tax_benefit_system_instance = system
    default_role = "member"
    default_calculation_period = CURRENT_YEAR
    default_input_period = CURRENT_YEAR

    def __init__(self, *args, **kwargs):
        start_instant: Annotated[str, "ISO date format YYYY-MM-DD"] = (
            kwargs.pop("start_instant", DEFAULT_START_DATE)
        )
        super().__init__(*args, **kwargs)

        reform = create_structural_reforms_from_parameters(
            self.tax_benefit_system.parameters, start_instant
        )
        if reform is not None:
            self.apply_reform(reform)

        # Labor supply responses

        employment_income = self.get_holder("employment_income")
        for known_period in employment_income.get_known_periods():
            array = employment_income.get_array(known_period)
            self.set_input("employment_income_before_lsr", known_period, array)
            employment_income.delete_arrays(known_period)

        self_employment_income = self.get_holder("self_employment_income")
        for known_period in employment_income.get_known_periods():
            array = self_employment_income.get_array(known_period)
            self.set_input(
                "self_employment_income_before_lsr", known_period, array
            )
            self_employment_income.delete_arrays(known_period)

        weekly_hours = self.get_holder("weekly_hours_worked")
        for known_period in weekly_hours.get_known_periods():
            array = weekly_hours.get_array(known_period)
            self.set_input(
                "weekly_hours_worked_before_lsr", known_period, array
            )
            weekly_hours.delete_arrays(known_period)

        # Capital gains responses

        cg_holder = self.get_holder("long_term_capital_gains")
        for known_period in cg_holder.get_known_periods():
            array = cg_holder.get_array(known_period)
            self.set_input(
                "long_term_capital_gains_before_response", known_period, array
            )
            cg_holder.delete_arrays(known_period)


class Microsimulation(CoreMicrosimulation):
    """
    A microsimulation of the tax-benefit system for the United States,
    defined against the base microsimulation class in the -core package.

    This simulation contains society-wide representative microdata, and is
    thus suitable for society-level impacts.

    Args:
        start_instant(str: ISO date format YYYY-MM-DD): Optional; The date
        at which the simulation begins; defaults to 2024-01-01; this is a
        temporary patch for structural reforms, and must be set to the start
        date of a structural reform parameter if it begins on a date other
        than the first day of the current year.
    """

    default_tax_benefit_system = CountryTaxBenefitSystem
    default_tax_benefit_system_instance = system
    default_dataset = DEFAULT_DATASET
    default_dataset_year = CURRENT_YEAR
    default_role = "member"
    default_calculation_period = CURRENT_YEAR
    default_input_period = CURRENT_YEAR

    def __init__(self, *args, **kwargs):
        start_instant: Annotated[str, "ISO date format YYYY-MM-DD"] = (
            kwargs.pop("start_instant", DEFAULT_START_DATE)
        )

        dataset = kwargs.get("dataset")
        if (
            dataset is not None
            and isinstance(dataset, str)
            and "cps_2023" in dataset
        ):
            self.default_input_period = 2023

        super().__init__(*args, **kwargs)

        reform = create_structural_reforms_from_parameters(
            self.tax_benefit_system.parameters, start_instant
        )
        if reform is not None:
            self.apply_reform(reform)

        # Labor supply responses

        employment_income = self.get_holder("employment_income")
        for known_period in employment_income.get_known_periods():
            array = employment_income.get_array(known_period)
            self.set_input("employment_income_before_lsr", known_period, array)
            employment_income.delete_arrays(known_period)

        self_employment_income = self.get_holder("self_employment_income")
        for known_period in self_employment_income.get_known_periods():
            array = self_employment_income.get_array(known_period)
            self.set_input(
                "self_employment_income_before_lsr", known_period, array
            )
            self_employment_income.delete_arrays(known_period)

        weekly_hours = self.get_holder("weekly_hours_worked")
        for known_period in weekly_hours.get_known_periods():
            array = weekly_hours.get_array(known_period)
            self.set_input(
                "weekly_hours_worked_before_lsr", known_period, array
            )
            weekly_hours.delete_arrays(known_period)

        # Capital gains responses

        cg_holder = self.get_holder("long_term_capital_gains")
        for known_period in cg_holder.get_known_periods():
            array = cg_holder.get_array(known_period)
            self.set_input(
                "long_term_capital_gains_before_response", known_period, array
            )
            cg_holder.delete_arrays(known_period)

        self.input_variables = [
            variable
            for variable in self.input_variables
            if variable
            not in [
                "employment_income",
                "self_employment_income",
                "weekly_hours_worked",
                "capital_gains",
            ]
        ] + [
            "employment_income_before_lsr",
            "self_employment_income_before_lsr",
            "weekly_hours_worked_before_lsr",
            "long_term_capital_gains_before_response",
        ]


class IndividualSim(CoreIndividualSim):  # Deprecated
    tax_benefit_system = CountryTaxBenefitSystem
    entities = {entity.key: entity for entity in entities}
    default_dataset = DEFAULT_DATASET
    default_roles = dict(
        tax_unit="member",
        spm_unit="member",
        household="member",
        family="member",
    )
    required_entities = [
        "tax_unit",
        "spm_unit",
        "household",
        "family",
    ]



from policyengine_us.model_api import *


class nm_low_income_comprehensive_tax_rebate_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico low income comprehensive tax rebate exemptions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://casetext.com/statute/new-mexico-statutes-1978/chapter-7-taxation/article-2-income-tax-general-provisions/section-7-2-14-low-income-comprehensive-tax-rebate?sort=relevance&type=regulation&tab=keyword&jxs=&resultsNav=false",
    )
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        """
        Quoting from N.M. Stat. § 7-2-14(C):
          For the purposes of this section, the total number of exemptions
          for which a tax rebate may be claimed or allowed is determined by
          adding the number of federal exemptions allowable for federal income
          tax purposes for each individual included in the return who is
          domiciled in New Mexico...
        """
        federal_exemptions = tax_unit("exemptions_count", period)
        """
          ...plus two additional exemptions for each individual domiciled
          in New Mexico included in the return who is sixty-five years of
          age or older...
        NB: The tax form shows that this and the blind exemption apply
            only to head and spouse, not dependents.
            https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf#page=58
        """
        p = parameters(
            period
        ).gov.states.nm.tax.income.rebates.low_income.exemptions
        aged_exemptions_head = p.aged.calc(tax_unit("age_head", period))
        aged_exemptions_spouse = p.aged.calc(tax_unit("age_spouse", period))
        """
          ...plus one additional exemption for each individual domiciled
          in New Mexico included in the return who, for federal income tax
          purposes, is blind...
        """
        blind_head_plus_spouse = add(
            tax_unit, period, ["blind_head", "blind_spouse"]
        )
        blind_exemptions = p.blind * blind_head_plus_spouse
        """
          ...plus one exemption for each minor child or stepchild of the
          resident who would be a dependent for federal income tax purposes
          if the public assistance contributing to the support of the child
          or stepchild was considered to have been contributed by the resident.
        NB: This section is not defined in the tax form, so we skip it.
        """
        return (
            federal_exemptions
            + aged_exemptions_head
            + aged_exemptions_spouse
            + blind_exemptions
        )



from policyengine_us.model_api import *


class household_weight(Variable):
    value_type = float
    entity = Household
    label = "Household weight"
    definition_period = YEAR
    uprating = "calibration.gov.census.populations.total"



from policyengine_us.model_api import *


class ny_ctc_post_2024_base(Variable):
    value_type = float
    entity = TaxUnit
    label = "New York CTC post-2024 base amount"
    documentation = (
        "Base New York CTC amount before phase-out under post-2024 rules"
    )
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c-1)
    defined_for = "ny_ctc_post_2024_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ny.tax.income.credits.ctc
        person = tax_unit.members
        age = person("age", period)

        # Post-2024 CTC rules (2025-2027)
        qualifies_for_federal_ctc = person("ctc_qualifying_child", period)
        qualifies = qualifies_for_federal_ctc & (age >= p.minimum_age)

        # Calculate credit amount by age using scale parameter
        credit_by_age = p.post_2024.amount.calc(age)
        qualifying_credit = qualifies * credit_by_age

        return tax_unit.sum(qualifying_credit)



from policyengine_us.model_api import *


class ny_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY CTC"
    documentation = "New York's Empire State Child Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c-1)
    defined_for = StateCode.NY
    adds = ["ny_ctc_post_2024", "ny_ctc_pre_2024"]



from policyengine_us.model_api import *


class de_personal_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware personal credit"
    unit = USD
    definition_period = YEAR
    reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf"
    defined_for = StateCode.DE

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.de.tax.income.credits
        exemptions_count = tax_unit("exemptions_count", period)
        return p.personal_credits.personal * exemptions_count



from policyengine_us.model_api import *


class pa_tax_forgiveness_rate(Variable):
    value_type = float
    entity = TaxUnit
    label = "PA tax forgiveness on eligibility income"
    unit = "/1"
    definition_period = YEAR
    reference = "https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2021/2021_pa-40in.pdf#page=39"
    defined_for = StateCode.PA

    def formula(tax_unit, period, parameters):
        eligibility_income = tax_unit("pa_eligibility_income", period)
        person = tax_unit.members
        is_child_dependent = person("is_child_dependent", period)
        child_dependents = tax_unit.sum(is_child_dependent)
        # filing status affects the base, where it doubles for married claimants
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values
        joint_separate = (filing_status == filing_statuses.JOINT) | (
            filing_status == filing_statuses.SEPARATE
        )
        base_multiplier = where(joint_separate, 2, 1)
        p = parameters(period).gov.states.pa.tax.income.forgiveness
        base = p.base * base_multiplier
        rate_per_dependent = p.dependent_rate
        eligibility_income_increment = base + (
            rate_per_dependent * child_dependents
        )
        excess = eligibility_income - eligibility_income_increment
        forgiveness_increment = p.rate_increment
        increments = np.ceil(excess / forgiveness_increment)
        percent = p.tax_back
        return min_(max_(1 - percent * increments, 0), 1)



from policyengine_us.model_api import *


def create_middle_class_tax_credit() -> Reform:
    class middle_class_tax_credit(Variable):
        value_type = float
        entity = TaxUnit
        label = "Middle Class Tax Credit"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://www.congress.gov/bill/116th-congress/senate-bill/4/text"
        )

        def formula(tax_unit, period, parameters):
            p = parameters(
                period
            ).gov.contrib.harris.lift.middle_class_tax_credit
            earned_income = add(tax_unit, period, ["earned_income"])
            pell_grant = add(tax_unit, period, ["pell_grant"])
            total_earned_income = earned_income + pell_grant
            filing_status = tax_unit("filing_status", period)
            joint = filing_status == filing_status.possible_values.JOINT
            cap = where(joint, p.cap * p.joint_multiplier, p.cap)
            capped_earned_income = min_(total_earned_income, cap)
            agi = tax_unit("adjusted_gross_income", period)

            excess = max_(0, agi - p.phase_out.start[filing_status])
            phase_out_rate = min_(1, excess / p.phase_out.width[filing_status])
            reduction = capped_earned_income * phase_out_rate
            age = tax_unit.members("age", period)
            age_eligible = age >= p.age_threshold
            eligible = tax_unit.any(age_eligible)
            return max_(0, capped_earned_income - reduction) * eligible

    class income_tax_refundable_credits(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "federal refundable income tax credits"
        unit = USD

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.irs.credits
            previous_credits = add(tax_unit, period, p.refundable)
            middle_class_credit = tax_unit("middle_class_tax_credit", period)
            return middle_class_credit + previous_credits

    class reform(Reform):
        def apply(self):
            self.update_variable(middle_class_tax_credit)
            self.update_variable(income_tax_refundable_credits)

    return reform


def create_middle_class_tax_credit_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_middle_class_tax_credit()

    p = parameters(period).gov.contrib.harris.lift.middle_class_tax_credit

    if p.in_effect:
        return create_middle_class_tax_credit()
    else:
        return None


middle_class_tax_credit = create_middle_class_tax_credit_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class ma_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA refundable credits"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/doc/2021-form-1-massachusetts-resident-income-tax-return/download"
    defined_for = StateCode.MA
    adds = "gov.states.ma.tax.income.credits.refundable"



from policyengine_us.model_api import *


class co_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado taxable income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CO

    def formula(tax_unit, period, parameters):
        fed_taxinc = tax_unit("taxable_income", period)
        additions = tax_unit("co_additions", period)
        subtractions = tax_unit("co_subtractions", period)
        return max_(0, fed_taxinc + additions - subtractions)



from policyengine_us.model_api import *


class oh_section_179_expense_add_back(Variable):
    value_type = float
    entity = Person
    label = "Ohio Section 179 Expense Add Back"
    definition_period = YEAR
    documentation = ""
    reference = ""
    defined_for = StateCode.OH
    # use federal variables if they are added later



from policyengine_us.model_api import *


class ia_alternate_tax_unit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa alternate tax calculated using worksheet"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=53"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=53"
    )
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        # compute alternate tax following worksheet in the instructions
        p = parameters(period).gov.states.ia.tax.income.alternate_tax
        # ... determine alternate tax deduction
        elder_head = tax_unit("age_head", period) >= p.elderly_age
        elder_spouse = tax_unit("age_spouse", period) >= p.elderly_age
        is_elder = elder_head | elder_spouse
        alt_ded = where(is_elder, p.deduction.elderly, p.deduction.nonelderly)
        # ... determine alternate tax amount
        alt_taxinc = max_(0, tax_unit("ia_modified_income", period) - alt_ded)
        return alt_taxinc * p.rate



from policyengine_us.model_api import *


class DCCCSPSCHEDULETYPE(Enum):
    FULL_TIME_TRADITIONAL = "Full-Time Traditional"
    PART_TIME_TRADITIONAL = "Part-Time Traditional"
    EXTENDED_DAY_FULL_TIME = "Extended Day Full-Time"
    EXTENDED_DAY_PART_TIME = "Extended Day Part-Time"
    FULL_TIME_NONTRADITIONAL = "Full-Time Nontraditional"
    PART_TIME_NONTRADITIONAL = "Part-Time Nontraditional"


class dc_ccsp_schedule_type(Variable):
    value_type = Enum
    entity = Person
    possible_values = DCCCSPSCHEDULETYPE
    default_value = DCCCSPSCHEDULETYPE.FULL_TIME_TRADITIONAL
    label = "DC Child Care Subsidy Program (CCSP) schedule type"
    definition_period = MONTH
    defined_for = StateCode.DC
    reference = "https://osse.dc.gov/sites/default/files/dc/sites/osse/publication/attachments/FY25%20Subsidy%20Reimbursement%20Rates%20English.pdf"



from policyengine_us.model_api import *


class ok_child_care_child_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Oklahoma Child Care/Child Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/511-Pkt-2021.pdf"
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-Pkt.pdf"
    )
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ok.tax.income.credits
        # determine AGI eligibility
        us_agi = tax_unit("adjusted_gross_income", period)
        agi_eligible = us_agi <= p.child.agi_limit
        # determine OK cdcc amount
        # Oklahoma matches the potential federal credit
        us_cdcc = tax_unit("cdcc_potential", period)
        ok_cdcc = us_cdcc * p.child.cdcc_fraction
        # determine OK ctc amount
        us_ctc = tax_unit("ctc_value", period)
        ok_ctc = us_ctc * p.child.ctc_fraction
        # determine prorated fraction
        ok_agi = tax_unit("ok_agi", period)
        # Compute OK AGI as a share of US AGI.
        # Use a mask rather than where to avoid a divide-by-zero warning.
        agi_ratio = np.zeros_like(us_agi)
        mask = us_agi != 0
        agi_ratio[mask] = ok_agi[mask] / us_agi[mask]
        prorate = min_(1, max_(0, agi_ratio))
        # receive greater of OK cdcc or OK ctc amounts prorated if AGI eligible
        return agi_eligible * prorate * max_(ok_cdcc, ok_ctc)



from policyengine_us.model_api import *


class ca_state_supplement_food_allowance(Variable):
    value_type = float
    entity = SPMUnit
    label = "California SSI state supplement food allowance"
    unit = USD
    definition_period = MONTH
    defined_for = "ca_state_supplement_food_allowance_eligible"
    reference = "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=WIC&sectionNum=12200"

    def formula(spm_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ca.cdss.state_supplement.payment_standard
        is_married = spm_unit("spm_unit_is_married", period)
        aged_disabled_count = spm_unit(
            "ca_state_supplement_aged_disabled_count", period
        )
        # Food allowance amount
        return where(
            is_married & (aged_disabled_count == 2),
            p.allowance.food.married,
            p.allowance.food.single,
        )



from policyengine_us.model_api import *


class taxsim_v18(Variable):
    value_type = float
    entity = TaxUnit
    label = "Taxable income in TAXSIM"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit("taxable_income", period)



from policyengine_us.model_api import *


class ma_scb_total_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Total income for the MA Senior Circuit Breaker"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/info-details/mass-general-laws-c62-ss-6"  # Part (k)
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        ma_gross_income = tax_unit(
            "ma_gross_income", period
        )  # The law specifies to start at AGI and re-add deducted capital losses. We instead start from gross income, on an equivalent path.
        scb = parameters(
            period
        ).gov.states.ma.tax.income.credits.senior_circuit_breaker
        disallowed_deductions = add(
            tax_unit, period, scb.income.disallowed_deductions
        )

        # Re-add some exemptions
        person = tax_unit.members
        tax = parameters(period).gov.states.ma.tax.income
        blind = person("is_blind", period)
        dependent = person("is_tax_unit_dependent", period)
        count_blind = tax_unit.sum(~dependent & blind)
        blind_exemption = tax.exemptions.blind * count_blind
        # (1C) and (2C): Aged exemptions.
        age = person("age", period)
        count_aged = tax_unit.sum(
            ~dependent & (age >= tax.exemptions.aged.age)
        )
        aged_exemption = tax.exemptions.aged.amount * count_aged
        # (3): Dependent exemptions.
        count_dependents = tax_unit("tax_unit_dependents", period)
        dependent_exemption = tax.exemptions.dependent * count_dependents
        exemptions = blind_exemption + aged_exemption + dependent_exemption

        return max_(0, ma_gross_income + disallowed_deductions - exemptions)



from policyengine_us.model_api import *


class self_employment_tax_ald_person(Variable):
    value_type = float
    entity = Person
    label = "Self-employment tax ALD deduction for each person"
    unit = USD
    documentation = "Personal above-the-line deduction for self-employment tax"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/164#f"

    def formula(person, period, parameters):
        self_employment_tax = person("self_employment_tax", period)
        percent_deductible = parameters(
            period
        ).gov.irs.ald.self_employment_tax.percent_deductible
        return self_employment_tax * percent_deductible



from policyengine_us.model_api import *


class co_social_security_subtraction_indv(Variable):
    value_type = float
    entity = Person
    label = "Colorado social security subtraction for eligible individuals"
    defined_for = "co_social_security_subtraction_indv_eligible"
    unit = USD
    reference = (
        "https://tax.colorado.gov/sites/tax/files/documents/DR0104AD_2022.pdf#page=1",
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=12",
        "https://law.justia.com/codes/colorado/2022/title-39/article-22/part-1/section-39-22-104/",
        # C.R.S. 39-22-104(4)(g)(III)
    )
    definition_period = YEAR

    def formula(person, period, parameters):
        p = parameters(period).gov.states.co.tax.income.subtractions.pension
        if not p.social_security_subtraction_available:
            return 0
        taxable_social_security = person("taxable_social_security", period)
        social_security_survivors = person("social_security_survivors", period)
        age = person("age", period)
        # Only head and spouse can claim this subtraction
        eligible_social_security_survivors = social_security_survivors
        eligible_taxable_social_security = taxable_social_security
        # If the filer is 65 or older, they can claim full subtarction
        # If the filer is between 65 and 55, they can claim a capped amount
        cap = p.cap.younger
        capped_middle_amount = min_(eligible_taxable_social_security, cap)
        # If the filer is under 55, they can claim a capped amount of ss survivors
        capped_younger_amount = min_(eligible_social_security_survivors, cap)
        return select(
            [
                age >= p.age_threshold.older,
                age >= p.age_threshold.younger,
            ],
            [taxable_social_security, capped_middle_amount],
            default=capped_younger_amount,
        )



from policyengine_us.model_api import *


class eitc_phase_out_start(Variable):
    value_type = float
    entity = TaxUnit
    label = "EITC phase-out start"
    unit = USD
    documentation = "Earnings above this level reduce EITC entitlement."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        count_children = tax_unit("eitc_child_count", period)
        eitc = parameters(period).gov.irs.credits.eitc
        is_joint = tax_unit("tax_unit_is_joint", period)
        joint_bonus = eitc.phase_out.joint_bonus.calc(count_children)
        phase_out_start = eitc.phase_out.start.calc(count_children)
        return phase_out_start + is_joint * joint_bonus



from policyengine_us.model_api import *


class tx_ccs_work_requirement_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Texas CCS work requirement eligible"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/texas/40-Tex-Admin-Code-SS-809-56"
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.tx.twc.ccs.work_requirements

        person = spm_unit.members
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        is_exempt = person("tx_ccs_work_exempt", period)
        work_hours = person("weekly_hours_worked", period.this_year)

        # Count non-exempt parents and calculate their total work hours
        is_non_exempt_parent = is_head_or_spouse & ~is_exempt
        num_non_exempt_parents = spm_unit.sum(is_non_exempt_parent)
        total_work_hours = spm_unit.sum(
            where(is_non_exempt_parent, work_hours, 0)
        )

        # Determine work hours requirement based on non-exempt parents
        # If 0-1 non-exempt parents: 25 hours
        # If 2+ non-exempt parents: 50 hours
        requirement = where(
            num_non_exempt_parents > 1, p.two_parent, p.single_parent
        )

        # Check each parent individually
        # A parent meets requirements if they are:
        # - Exempt from work, OR
        # - Part of a household that meets the work hour requirement
        household_meets_work_requirement = total_work_hours >= requirement
        parent_meets_requirement = is_exempt | household_meets_work_requirement

        # SPM unit is eligible if ALL parents meet requirements
        # (Non-parents automatically pass)
        person_eligible = parent_meets_requirement | ~is_head_or_spouse
        return spm_unit.all(person_eligible)



from policyengine_us.model_api import *


class chip(Variable):
    value_type = float
    entity = Person
    label = "CHIP"
    unit = USD
    definition_period = YEAR
    reference = "https://www.macpac.gov/publication/chip-spending-by-state/"
    defined_for = "is_chip_eligible"
    adds = ["per_capita_chip"]



from policyengine_us.model_api import *


def create_halve_joint_eitc_phase_out_rate() -> Reform:
    class eitc_phase_out_rate(Variable):
        value_type = float
        entity = TaxUnit
        label = "EITC phase-out rate"
        unit = USD
        documentation = "Percentage of earnings above the phase-out threshold that reduce the EITC."
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            eitc = parameters(period).gov.irs.credits.eitc
            num_children = tax_unit("eitc_child_count", period)
            filing_status = tax_unit("filing_status", period)
            joint = filing_status == filing_status.possible_values.JOINT
            base_rate = eitc.phase_out.rate.calc(num_children)
            return where(joint, base_rate / 2, base_rate)

    class reform(Reform):
        def apply(self):
            self.update_variable(eitc_phase_out_rate)

    return reform


def create_halve_joint_eitc_phase_out_rate_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_halve_joint_eitc_phase_out_rate()

    p = parameters(period).gov.contrib.joint_eitc

    if p.in_effect:
        return create_halve_joint_eitc_phase_out_rate()
    else:
        return None


halve_joint_eitc_phase_out_rate = (
    create_halve_joint_eitc_phase_out_rate_reform(None, None, bypass=True)
)



from policyengine_us.model_api import *


class co_federal_ctc_child_individual_maximum(Variable):
    value_type = float
    entity = Person
    label = "CTC maximum amount (child) replicated to account for the Colorado state CTC child eligibility"
    unit = USD
    documentation = "The CTC entitlement in respect of this person as a child."
    definition_period = YEAR
    reference = (
        # C.R.S. 39-22-129. Child tax credit - legislative declaration - definitions.
        "https://casetext.com/statute/colorado-revised-statutes/title-39-taxation/specific-taxes/income-tax/article-22-income-tax/part-1-general/section-39-22-129-child-tax-credit-legislative-declaration-definitions-repeal",
        # 2022 Colorado Child Tax Credit
        "https://tax.colorado.gov/sites/tax/files/documents/DR_0104CN_2022.pdf#page=1",
        # Colorado Individual Income Tax Filing Guide - Instructions for Select Credits from the DR 0104CR - Line 1 Child Tax Credit
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=16",
    )
    defined_for = StateCode.CO

    def formula(person, period, parameters):
        age = person("age", period)
        base_amount = parameters(period).gov.irs.credits.ctc.amount.base
        is_dependent = person("is_tax_unit_dependent", period)
        p = parameters(period).gov.states.co.tax.income.credits.ctc
        eligible = is_dependent & (age < p.age_threshold)
        return base_amount.calc(age) * eligible



from policyengine_us.model_api import *


class ms_taxable_income_indiv(Variable):
    value_type = float
    entity = Person
    label = "Mississippi taxable income when married couple file separately"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100221.pdf#page=13",
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80105228.pdf",  # Line 38 - 49,
    )
    defined_for = StateCode.MS

    def formula(person, period, parameters):
        pre_deductions_agi = person(
            "ms_pre_deductions_taxable_income_indiv", period
        )
        deductions = person("ms_deductions_indiv", period)
        return max_(pre_deductions_agi - deductions, 0)



from policyengine_us.model_api import *


class is_ssi_disabled(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    documentation = "Indicates whether a person is disabled for the Supplemental Security Income program"
    label = "SSI disabled"
    reference = "https://www.law.cornell.edu/uscode/text/42/1382c#a_3_A"

    def formula(person, period, parameters):
        # Classify person as disabled for SSI if:
        # 1) They reported receiving SSI due to presumed disability status (being neither aged nor blind); or
        # 2) They are disabled.
        # Earnings in excess of the Substantial Gainful Activity threshold disqualify from either case.
        # First check case 1: SSI reported and neither aged nor blind.
        aged = person("is_ssi_aged", period)
        blind = person("is_blind", period)
        reported_receipt = person("ssi_reported", period) > 0
        reported_disabled_ssi = reported_receipt & ~aged & ~blind
        is_disabled = person("is_disabled", period)
        engaged_in_ssa = person("ssi_engaged_in_sga", period)
        return (reported_disabled_ssi | is_disabled) & ~engaged_in_ssa



from policyengine_us.model_api import *


class wv_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia income tax before refundable tax credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.WV

    def formula(tax_unit, period, parameters):
        tax_before_credits = tax_unit(
            "wv_income_tax_before_non_refundable_credits", period
        )
        credits = tax_unit("wv_non_refundable_credits", period)
        return max_(tax_before_credits - credits, 0)



from policyengine_us.model_api import *


class ma_mbta_income_eligible_reduced_fare_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for the Massachusetts Bay Transportation Authority income-eligible reduced fare program"
    definition_period = YEAR
    defined_for = "ma_mbta_enrolled_in_applicable_programs"
    reference = "https://www.mbta.com/fares/reduced/income-eligible"

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.ma.dot.mbta.income_eligible_reduced_fares.age_threshold
        age = person("age", period)
        return p.calc(age)



from policyengine_us.model_api import *


class alimony_expense_ald(Variable):
    value_type = float
    entity = TaxUnit
    label = "Alimony expense ALD"
    unit = USD
    documentation = (
        "Above-the-line deduction from gross income for alimony expenses."
    )
    definition_period = YEAR
    reference = "https://www.irs.gov/taxtopics/tc452"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        divorce_year = person("divorce_year", period)
        alimony_expense = person("alimony_expense", period)
        p = parameters(period).gov.irs.ald.alimony_expense
        eligible_person = p.divorce_year_threshold.calc(divorce_year)
        return tax_unit.sum(alimony_expense * eligible_person)



from policyengine_us.model_api import *


class nyc_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "NYC EITC"
    unit = USD
    definition_period = YEAR
    defined_for = "in_nyc"

    def formula(tax_unit, period, parameters):
        # First get their NYAGI and EITC.
        ny_agi = tax_unit("ny_agi", period)
        federal_eitc = tax_unit("eitc", period)

        # Then get the EITC part of the parameter tree.
        p = parameters(period).gov.local.ny.nyc.tax.income.credits.eitc

        # Calculate the EITC before the reduction.
        percentage_pre_reduction = p.percent.calc(ny_agi)

        # Calculate the reduction (if applicable).
        threshold = p.reduction_threshold.calc(ny_agi)
        excess = max_(0, ny_agi - threshold)
        pct_point_reduction_if_applicable = p.percent_reduction * excess
        pct_point_reduction = where(
            threshold > 0, pct_point_reduction_if_applicable, 0
        )

        # Percentage is nonnegative because the last bracket is a flat 10%.
        percentage = percentage_pre_reduction - pct_point_reduction

        # Return the net EITC.
        return federal_eitc * percentage



from policyengine_us.model_api import *


class mt_medical_expense_deduction_joint(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Montana medical expense deduction when married couples are filing jointly"
    reference = (
        "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/05/Montana-Idividiual-Income-Tax-Return-Form-2-2022v6.2.pdf#page=7",
        "https://law.justia.com/codes/montana/2022/title-15/chapter-30/part-21/section-15-30-2131/",
        # MT Code § 15-30-2131 (2022) (1)(g)(i)
    )
    unit = USD
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        expense = add(
            person.tax_unit, period, ["medical_out_of_pocket_expenses"]
        )
        p = parameters(period).gov.irs.deductions.itemized.medical
        # Law does not define Montana AGI as the cap.
        # Tax form points to page 1, line 14, which is Montana AGI.
        medical_floor = p.floor * add(
            person.tax_unit, period, ["mt_agi_indiv"]
        )
        is_head = person("is_tax_unit_head", period)
        return is_head * max_(0, expense - medical_floor)



from policyengine_us.model_api import *


class wv_senior_citizen_disability_deduction_total_modifications(Variable):
    value_type = float
    entity = Person
    label = "West Virginia total modifications for the senior citizen or disability deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.WV

    adds = "gov.states.wv.tax.income.subtractions.senior_citizen_disability_deduction.modification_sources"



from policyengine_us.model_api import *


class ma_child_and_family_credit_or_dependent_care_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA dependent or dependent care credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.mass.gov/info-details/mass-general-laws-c62-ss-6"  # (x-y)
    )
    defined_for = StateCode.MA
    documentation = "The CFTC replaces the Dependent Care Tax Credit and the Household Dependent Tax Credit with Child and Family Tax Credit for tax years beginning on or after January 1, 2023."

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ma.tax.income.credits.dependent_care
        # MA taxpayers can only take either the dependent credit or the
        # dependent care credit.
        dependent_credit = tax_unit("ma_child_and_family_credit", period)
        if p.in_effect:
            return max_(
                dependent_credit,
                tax_unit("ma_dependent_care_credit", period),
            )
        return dependent_credit



from policyengine_us.model_api import *


class is_permanently_disabled_veteran(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Permanently disabled veteran"



from policyengine_us.model_api import *


class in_eitc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Indiana earned income tax credit eligibility status"
    unit = USD
    definition_period = YEAR
    reference = "https://iga.in.gov/laws/2021/ic/titles/6#6-3.1-21"
    defined_for = StateCode.IN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states["in"].tax.income
        # check federal eitc receipt
        gets_federal_eitc = tax_unit("eitc", period) > 0
        if not p.credits.earned_income.decoupled:
            return gets_federal_eitc
        # if Indiana EITC is decoupled from federal EITC
        # ... check separate filing status
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        # ... check age eligibility for childless taxpayers
        is_childless = tax_unit("eitc_child_count", period) == 0
        min_age = p.credits.earned_income.childless.min_age
        max_age = p.credits.earned_income.childless.max_age
        age_head = tax_unit("age_head", period)
        age_spouse = tax_unit("age_spouse", period)
        head_age_eligible = (age_head >= min_age) & (age_head <= max_age)
        spouse_age_eligible = (age_spouse >= min_age) & (age_spouse <= max_age)
        married = filing_status.possible_values.JOINT
        age_eligible = where(
            married, head_age_eligible | spouse_age_eligible, head_age_eligible
        )
        childless_age_eligible = where(is_childless, age_eligible, True)
        # ... check investment income eligibility
        invinc = tax_unit("eitc_relevant_investment_income", period)
        invinc_limit = p.credits.earned_income.investment_income_limit
        invinc_eligible = invinc <= invinc_limit
        # ... determine Indiana EITC eligibility status
        in_eligible = ~separate & childless_age_eligible & invinc_eligible
        return gets_federal_eitc & in_eligible



from policyengine_us.model_api import *


class md_tanf_gross_earned_income_deduction(Variable):
    value_type = float
    entity = SPMUnit
    label = "Maryland TANF earned income deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MD

    def formula(spm_unit, period, parameters):
        # Get TANF enrollment status.
        is_tanf_enrolled = spm_unit("is_tanf_enrolled", period)
        # Get earned income for the SPM unit.
        earned_income = add(spm_unit, period, ["earned_income"])
        # Determine if the SPM unit has any self-employment income.
        self_employment_income = add(
            spm_unit, period, ["self_employment_income"]
        )
        has_self_employment_income = self_employment_income > 0
        # Get the policy parameters.
        p = parameters(period).gov.states.md.tanf.income.deductions.earned
        percent = select(
            # First arg: list of conditions
            [~is_tanf_enrolled, has_self_employment_income],
            # Second arg: list of values to return if the corresponding condition is True
            [p.new, p.self_employed],
            # Third arg: default value to return if none of the conditions are True
            default=p.not_self_employed,
        )
        # Multiply earned income by percent deduction.
        return earned_income * percent



from policyengine_us.model_api import *


class refundable_american_opportunity_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Refundable American Opportunity Credit"
    unit = USD
    documentation = (
        "Value of the refundable portion of the American Opportunity Credit"
    )
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25A#i"

    def formula(tax_unit, period, parameters):
        aoc = parameters(
            period
        ).gov.irs.credits.education.american_opportunity_credit
        return aoc.refundability * tax_unit(
            "american_opportunity_credit", period
        )



from policyengine_us.model_api import *


class ctc_individual_maximum(Variable):
    value_type = float
    entity = Person
    label = "CTC individual amount maximum"
    unit = USD
    documentation = (
        "The Child Tax Credit entitlement in respect of this person."
    )
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/24#a",
        "https://www.law.cornell.edu/uscode/text/26/24#h",
        "https://www.law.cornell.edu/uscode/text/26/24#i",
    )
    adds = ["ctc_child_individual_maximum", "ctc_adult_individual_maximum"]



from policyengine_us.model_api import *


class il_schedule_m_additions(Variable):
    value_type = float
    entity = TaxUnit
    label = "IL Schedule M additions"
    unit = USD
    definition_period = YEAR
    reference = "https://www2.illinois.gov/rev/forms/incometax/Documents/currentyear/individual/il-1040-schedule-m.pdf"
    defined_for = StateCode.IL



from policyengine_us.model_api import *


class il_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "IL EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://www2.illinois.gov/rev/programs/EIC/Pages/default.aspx"
    defined_for = StateCode.IL

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        match = parameters(period).gov.states.il.tax.income.credits.eitc.match
        return federal_eitc * match



from policyengine_us.model_api import *


class ar_additional_tax_credit_for_qualified_individuals_person(Variable):
    value_type = float
    entity = Person
    label = "Arkansas additional tax credit for qualified individuals for each individual"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        filing_separately = person.tax_unit("ar_files_separately", period)
        income_joint = person("ar_taxable_income_joint", period)
        income_indiv = person("ar_taxable_income_indiv", period)
        # When filing separately, the credit is calculated based on individual income
        income = where(
            filing_separately, income_indiv, person.tax_unit.sum(income_joint)
        )
        p = parameters(
            period
        ).gov.states.ar.tax.income.credits.additional_tax_credit_for_qualified_individuals
        filing_status = person.tax_unit("filing_status", period)
        joint = filing_status == filing_status.possible_values.JOINT
        filing_jointly = joint & ~filing_separately
        multiplier = where(filing_jointly, p.joint_multiplier, 1)
        max_amount = where(
            filing_jointly, p.max_amount * multiplier, p.max_amount
        )
        excess = max_(income - p.reduction.start, 0)
        increments = np.ceil(excess / p.reduction.increment)
        reduction_amount = where(
            filing_jointly,
            p.reduction.amount * multiplier,
            p.reduction.amount,
        )
        total_reduction_amount = increments * reduction_amount
        # Attribute the maximum amount to each spouse equally when married filing jointly
        divisor = where(filing_jointly, 2, 1)
        total_credit = max_(max_amount - total_reduction_amount, 0) / divisor
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        return total_credit * head_or_spouse



from policyengine_us.model_api import *


class co_ccap_child_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Child eligibility for Colorado Child Care Assistance Program"
    reference = "https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11042&fileName=8%20CCR%201403-1#page=6"
    definition_period = MONTH
    defined_for = StateCode.CO

    def formula(person, period, parameters):
        year = period.start.year
        if period.start.month >= 10:
            instant_str = f"{year}-10-01"
        else:
            instant_str = f"{year - 1}-10-01"
        p = parameters(instant_str).gov.states.co.ccap
        # child < 13 or disabled child < 19 to be eligible
        disabled = person("is_disabled", period.this_year)
        age_limit = where(disabled, p.disabled_age_limit, p.age_limit)
        age_eligible = person("age", period.this_year) < age_limit
        return age_eligible & person("is_tax_unit_dependent", period.this_year)



from policyengine_us.model_api import *


class tx_tanf(Variable):
    value_type = float
    entity = SPMUnit
    label = "Texas Temporary Assistance for Needy Families (TANF)"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-2410-general-policy",
    )
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        # Compare monthly TANF benefit with monthly equivalent of OTTANF
        # Return whichever is higher for the household

        # Regular TANF: monthly benefit
        regular_tanf = spm_unit("tx_regular_tanf", period)

        # OTTANF: one-time payment divided by 12 for monthly equivalent
        ottanf = spm_unit("tx_ottanf", period)

        return max_(regular_tanf, ottanf)



from policyengine_us.model_api import *


class ma_part_b_taxable_income_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Part B taxable income deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/info-details/mass-general-laws-c62-ss-3"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        tax = parameters(period).gov.states.ma.tax.income
        # (B)(a)(3): Taxes for retirement programs.
        # NB: The law only mentions FICA and FRRA, but mass.gov includes SECA.
        # https://www.mass.gov/service-details/learn-about-business-and-professional-income
        person = tax_unit.members
        SS_MEDICARE_VARIABLES = [
            "employee_social_security_tax",
            "employee_medicare_tax",
            "self_employment_social_security_tax",
            "self_employment_medicare_tax",
        ]
        fica_person = add(person, period, SS_MEDICARE_VARIABLES)
        fica_head = min_(
            tax.deductions.public_retirement_contributions,
            fica_person * person("is_tax_unit_head", period),
        )
        fica_spouse = min_(
            tax.deductions.public_retirement_contributions,
            fica_person * person("is_tax_unit_spouse", period),
        )
        fica = tax_unit.sum(fica_head) + tax_unit.sum(fica_spouse)
        # Bank interest deduction.
        filing_status = tax_unit("filing_status", period)
        if tax.exemptions.interest.in_effect:
            bank_interest = add(tax_unit, period, ["taxable_interest_income"])
            bank_interest_deduction = min_(
                tax.exemptions.interest.amount[filing_status],
                bank_interest,
            )
        else:
            bank_interest_deduction = 0
        # (B)(a)(9): Rent deduction.
        rent = add(tax_unit, period, ["rent"])
        rent_deduction = tax.deductions.rent.share * rent
        rent_deduction = min_(
            rent_deduction,
            tax.deductions.rent.cap[filing_status],
        )
        # (B)(a)(13): Charitable contributions deduction.
        charitable_deduction = tax_unit("charitable_deduction", period)
        return (
            fica
            + bank_interest_deduction
            + rent_deduction
            + charitable_deduction
        )



from policyengine_us.model_api import *


class hud_max_subsidy(Variable):
    value_type = float
    entity = SPMUnit
    label = "HUD max subsidy"
    unit = USD
    documentation = "Max subsidy for HUD programs"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1437a#b_2_B_5"

    def formula(spm_unit, period, parameters):
        ttp = spm_unit("hud_ttp", period)
        pha_payment_standard = spm_unit.household(
            "pha_payment_standard", period
        )
        return max_(0, pha_payment_standard - ttp)



from policyengine_us.model_api import *


class dc_snap_temporary_local_benefit(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    documentation = "DC temporary SNAP benefit amount"
    label = "DC temporary local SNAP benefit amount"
    reference = (
        "https://dhs.dc.gov/page/give-snap-raise-heres-what-expect",
        "https://code.dccouncil.gov/us/dc/council/laws/24-301",
    )
    unit = USD
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.dc.tax.income.snap.temporary_local_benefit

        max_allotments = spm_unit("snap_max_allotment", period)
        eligible = spm_unit("is_snap_eligible", period)
        return max_allotments * p.rate * eligible



from policyengine_us.model_api import *


class nm_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico income deductions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        itemized_on_federal_return = tax_unit("tax_unit_itemizes", period)
        itm_ded = tax_unit("nm_itemized_deductions", period)
        standard_ded = tax_unit("standard_deduction", period)
        # Tax filer is required to itemize deductions if they itemize on their federal return as per:
        # https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
        itemized_or_standard = where(
            itemized_on_federal_return,
            itm_ded,
            standard_ded,
        )
        OTHER_DEDUCTIONS = [
            "nm_medical_care_expense_deduction",
            "nm_deduction_for_certain_dependents",
            "nm_net_capital_gains_deduction",
        ]
        other_deductions = add(tax_unit, period, OTHER_DEDUCTIONS)
        return itemized_or_standard + other_deductions



from policyengine_us.model_api import *


class ca_cvrp_vehicle_rebate_amount(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "CVRP rebate for purchased vehicle"
    unit = USD
    documentation = "Rebate value for a purchased vehicle under the California Clean Vehicle Rebate Project (CVRP)"
    reference = "https://cleanvehiclerebate.org/en/eligible-vehicles"
    defined_for = StateCode.CA



from policyengine_us.model_api import *


class il_tanf_countable_gross_earned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Illinois Temporary Assistance for Needy Families (TANF) countable gross earned income"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.IL

    def formula(spm_unit, period, parameters):
        person = spm_unit.members
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        gross_earned_income = person("il_tanf_gross_earned_income", period)
        return spm_unit.sum(is_head_or_spouse * gross_earned_income)



from policyengine_us.model_api import *


class nc_scca_maximum_payment(Variable):
    value_type = float
    entity = SPMUnit
    label = "North Carolina Subsidized Child Care Assistance Program maximum payment"
    unit = USD
    definition_period = MONTH
    reference = "https://www.publichealthlawcenter.org/sites/default/files/NC%20Chapter%2010%20Subsidized%20Child%20Care%20Rules.pdf#page=2"
    defined_for = "nc_scca_entry_eligible"

    def formula(spm_unit, period, parameters):
        total_market_rate = add(spm_unit, period, ["nc_scca_market_rate"])

        parent_fee = spm_unit("nc_scca_parent_fee", period)
        return max_(total_market_rate - parent_fee, 0)



from policyengine_us.model_api import *


class az_charitable_contributions_to_qualifying_charitable_organizations(
    Variable
):
    value_type = float
    entity = TaxUnit
    label = "Charitable contributions to qualifying charitable organizations in Arizona"
    definition_period = YEAR
    defined_for = StateCode.AZ
    reference = (
        "https://law.justia.com/codes/arizona/2022/title-43/section-43-1088/"
    )



from policyengine_us.model_api import *


class social_security_dependents(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Social Security dependents benefits"
    unit = USD
    # uprating = "gov.ssa.uprating"



from policyengine_us.model_api import *


class us_govt_interest_person(Variable):
    value_type = float
    entity = Person
    label = "Interest on U.S. government obligations for each person"
    unit = USD
    definition_period = YEAR
    documentation = "Interest on U.S. government obligations such as U.S. savings bonds, U.S. Treasury bills, and U.S. government certificates, allocated to each person."



from policyengine_us.model_api import *


class co_denver_homeowner_property_tax_relief(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Denver Property Tax Relief for homeowners"
    defined_for = "co_denver_property_tax_relief_homeowner_eligible"
    definition_period = YEAR
    reference = "https://denvergov.org/files/content/public/v/37/government/agencies-departments-offices/agencies-departments-offices-directory/denver-human-services/be-supported/additional-assistance/property-tax-relief/denver-property-tax-relief-program-year-2021-rules.pdf"

    adds = ["gov.local.co.denver.dhs.property_tax_relief.amount.homeowner"]



from policyengine_us.model_api import *


class ma_ccfa_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Massachusetts Child Care Financial Assistance (CCFA)"
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = "https://www.mass.gov/doc/eecs-financial-assistance-procedures-manual-february-1-2022/download"

    def formula(spm_unit, period, parameters):
        # Check all eligibility criteria
        has_eligible_child = (
            add(spm_unit, period, ["ma_ccfa_eligible_child"]) > 0
        )

        # TAFDC recipients have categorical eligibility Section 10.05(3)(a)&(b)
        # https://regulations.justia.com/states/massachusetts/606-cmr/title-606-cmr-10-00/section-10-05/
        tafdc_eligible = spm_unit("ma_tafdc_eligible", period)

        # Regular eligibility criteria
        income_eligible = spm_unit("ma_ccfa_income_eligible", period)
        asset_eligible = spm_unit("ma_ccfa_asset_eligible", period)
        activity_eligible = spm_unit("ma_ccfa_activity_eligible", period)

        regular_eligible = (
            has_eligible_child
            & income_eligible
            & asset_eligible
            & activity_eligible
        )

        return regular_eligible | tafdc_eligible



from policyengine_us.model_api import *


class pr_personal_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Puerto Rico personal exemption"
    reference = "https://hacienda.pr.gov/sites/default/files/individuals_2024_rev._jul_12_24_9-30-24_informative.pdf#page=2"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.PR

    def formula(tax_unit, period, parameters):
        # line 7
        p = parameters(
            period
        ).gov.territories.pr.tax.income.taxable_income.exemptions
        filing_status = tax_unit("filing_status", period)
        return p.personal[filing_status]



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant


def tax_employer_medicare_tax_reform() -> Reform:
    """
    General workflow for creating reform:
    1. Make a boolean for the reform under parameters/contrib/{reform name}
    called in_effect.
    2. Create new variables if necessary.
    3. Redefine rule for calculating existing variable (in this case social
    security taxation).
    3a. Add an init file to initialize the create_{reform name}_reform function.
    4. Add this reform to reforms/reforms.py (see previous pull requests for
    syntax).
    5. Add a unit test under tests/contrib/{reform name}
    if the reform is not just modifying a parameter
    5a. Add unit test for new variables under
    tests/baseline/{new variable's path}

    Specific to this reform:
    We want to modify how an existing variable is calculated.
    In this case it's "irs_gross_income", as in:
    "policyengine_us/variables/gov/irs/income/taxable_income/
    adjusted_gross_income/irs_gross_income/irs_gross_income.py"
    We amend irs_gross_income because the employer-side contributions
    will go into taxable income...
    but taxable income is determined at the tax unit level,
    by summing up all persons' gross income
    and deducting tax unit's deductions. But since payroll is calculated at
    the personal level,
    we want to add each person's employer contribution to each person, thus
    we add it to gross income.

    irs_gross_income refers to "parameters/gov/irs/gross_income/sources"
    to determine which sources of income get added into the variable.
    We modify that list by adding the two variables we created for this
    reform: they are employer's payroll taxes for
    (i) social security and (ii) medicare.
    The new variabes are under "variables/gov/irs/tax/payroll
    /social_security/employer_social_security_tax.py"
    and "policyengine_us/variables/gov/irs/tax/payroll/medicare/
    employer_medicare_tax.py"
    """

    class irs_gross_income(Variable):
        value_type = float
        entity = Person
        label = "IRS gross income"
        definition_period = YEAR
        unit = USD

        def formula(person, period, parameters):
            sources = parameters(period).gov.irs.gross_income.sources
            total = 0
            not_dependent = ~person("is_tax_unit_dependent", period)
            for source in sources:
                # Add positive values only - losses are deducted later.
                total += not_dependent * max_(0, add(person, period, [source]))
            return total + person("employer_medicare_tax", period)

    # Create a reform object applies the method
    # It inherits the Reform class
    class reform(Reform):
        def apply(self):
            self.update_variable(irs_gross_income)

    return reform


def create_tax_employer_medicare_tax_reform(
    parameters, period, bypass: bool = False
):
    # Create a create_{reform name} function that initializes the reform object
    # There are two sufficient conditions for this function to return
    # the reform

    # 1. If bypass is set to true
    if bypass is True:
        return tax_employer_medicare_tax_reform()

    # 2. If boolean in in_effect.yaml is set to true
    parameter = parameters.gov.contrib.crfb.tax_employer_medicare_tax
    current_period = period_(period)
    reform_active = False

    for i in range(5):
        if parameter(current_period).in_effect:
            # If in any of the next five years, the boolean is true,
            # set the boolean reform_active to true, and stop the check,
            # i.e., assume the reform is active in all subsequent years.
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    # if the loop set reform_active to true, return the reform.
    if reform_active:
        return tax_employer_medicare_tax_reform()
    else:
        return None


# Create a reform object to by setting bypass to true,
# for the purpose of running tests
tax_employer_medicare_tax_reform_object = (
    create_tax_employer_medicare_tax_reform(None, None, bypass=True)
)



from policyengine_us.model_api import *


class ar_itemized_deductions_joint(Variable):
    value_type = float
    entity = Person
    label = "Arkansas itemized deductions when married filing jointly"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR3_ItemizedDeduction.pdf"
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        # Arkansas does not tie itemization choice to federal choice.
        deductions = parameters(
            period
        ).gov.states.ar.tax.income.deductions.itemized.sources.joint
        unit_deds = add(person.tax_unit, period, deductions)
        is_head = person("is_tax_unit_head", period)
        return unit_deds * is_head



from policyengine_us.model_api import *


class mn_cdcc_phase_out(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota child and dependent care expense credit phase out"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-02/m1cd_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1cd_22_0.pdf"
    )
    defined_for = "mn_cdcc_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.credits.cdcc
        dep_count = tax_unit("mn_cdcc_dependent_count", period)
        # calculate pre-phaseout credit amount
        agi = tax_unit("adjusted_gross_income", period)
        # calculate post-phaseout credit amount
        excess_agi = max_(0, agi - p.phaseout_threshold)
        reduced_agi = excess_agi * p.phaseout_rate
        reduced_base_amount = p.reduced_base_amount.calc(dep_count)
        return max_(0, reduced_base_amount - reduced_agi)



from policyengine_us.model_api import *


class tx_tanf_budgetary_needs_test(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets Texas TANF budgetary needs test"
    definition_period = MONTH
    reference = "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1340-income-limits"
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        # Budgetary needs test applies to applicants (not receiving TANF in last 4 months)
        # Household passes if budgetary needs > income

        budgetary_needs = spm_unit("tx_tanf_budgetary_needs", period)
        income_for_test = spm_unit(
            "tx_tanf_income_for_budgetary_needs_test", period
        )

        return budgetary_needs > income_for_test
